////////////////////////////////////////////////////////////////
///
/// File routines
///
/////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include "MyTypes.h"
#include "error.h"

////////////////////////////////////////////////////
///
/// Ouvre un fichier
///
////////////////////////////////////////////////////
FILE *FILE_Open(char *p_pFilename, char *p_pMode)
{
	FILE	*pFile;

	if ((pFile = fopen(p_pFilename, p_pMode)) == NULL)
	{
		Error("Could not open file %s\n", p_pFilename);
	}

	return (pFile);
}

////////////////////////////////////////////////////
///
/// Ferme un fichier
///
////////////////////////////////////////////////////
void FILE_Close(FILE **p_pFile)
{
	if (*p_pFile != NULL)
	{
		fclose(*p_pFile);
		*p_pFile = NULL;
	}
}

////////////////////////////////////////////////////
///
/// Ecrit dans un fichier. Si erreur, affiche un  message et ferme le fichier
///
////////////////////////////////////////////////////
u32 FILE_Write(FILE **p_pFile, u8 *p_pMem, u32 p_size, char *p_pFilename)
{
	if (fwrite(p_pMem, 1, p_size, *p_pFile) != p_size)
	{
		if (p_pFilename != NULL)
		{
			Error("Error writing file %s\n", p_pFilename);
		}
		else
		{
			Error("Write error\n");
		}

		FILE_Close(p_pFile);

		return (FALSE);
	}

	return (TRUE);
}

////////////////////////////////////////////////////
///
/// Lit un fichier. Si erreur, affiche un  message et ferme le fichier
///
////////////////////////////////////////////////////
u32 FILE_Read(FILE **p_pFile, u8 *p_pMem, u32 p_size, char *p_pFilename)
{
	if (fread(p_pMem, 1, p_size, *p_pFile) != p_size)
	{
		if (p_pFilename != NULL)
		{
			Error("Error reading file %s\n", p_pFilename);
		}
		else
		{
			Error("Read error\n");
		}

		FILE_Close(p_pFile);

		return (FALSE);
	}

	return (TRUE);
}
