#include "StdAfx.h"
#include <stdio.h>
#include <commdlg.h>
#include <shlobj.h>
#include <ObjBase.h>
#include "MyTypes.h"
#include "bitmap.h"
#include "winstuf.h"
#include "error.h"

TWIN	g_Win;

//////////////////////////////////////////////////////////////////////////
/////
///// Le gros boxon pour retrouver le nom d'un rpertoire (pourquoi faire simple...)
/////
//////////////////////////////////////////////////////////////////////////

// Macros for interface casts

// Retrieves the UIObject interface for the specified full PIDL
STDAPI SHGetUIObjectFromFullPIDL(LPCITEMIDLIST pidl, HWND hwnd, REFIID riid, void **ppv)
{
    LPCITEMIDLIST pidlChild;
    IShellFolder* psf;

    *ppv = NULL;

    HRESULT hr = SHBindToParent(pidl, IID_IShellFolder, (void **)&psf, &pidlChild);
    if (SUCCEEDED(hr))
    {
        hr = psf->GetUIObjectOf(hwnd, 1, &pidlChild, riid, NULL, ppv);
        psf->Release();
    }
    return hr;
}
 
#define ILSkip(pidl, cb)       ((LPITEMIDLIST)(((BYTE*)(pidl))+cb))
#define ILNext(pidl)           ILSkip(pidl, (pidl)->mkid.cb)
 
HRESULT SHILClone(LPCITEMIDLIST pidl, LPITEMIDLIST *ppidl)
{
    DWORD cbTotal = 0;

    if (pidl) 
    {
        cbTotal += sizeof(pidl->mkid.cb);
        while (pidl->mkid.cb) 
        {
            cbTotal += pidl->mkid.cb;
            pidl = ILNext(pidl);
        }
    }
    
    *ppidl = (LPITEMIDLIST)CoTaskMemAlloc(cbTotal);
    
    if (*ppidl)
        CopyMemory(*ppidl, pidl, cbTotal);
 
    return  *ppidl ? S_OK: E_OUTOFMEMORY;
}
 
// Get the target PIDL for a folder PIDL. This deals with cases where a folder
// is an alias to a real folder, folder shortcuts, etc.
STDAPI SHGetTargetFolderIDList(LPCITEMIDLIST pidlFolder, LPITEMIDLIST *ppidl)
{
    IShellLink *psl;
	
    *ppidl = NULL;
    
    HRESULT hr = SHGetUIObjectFromFullPIDL(pidlFolder, NULL, IID_IShellLink, (void **)&psl);
    
    if (SUCCEEDED(hr))
    {
        hr = psl->GetIDList(ppidl);
        psl->Release();
    }
    
    // It's not a folder shortcut so get the PIDL normally.
    if (FAILED(hr))
        hr = SHILClone(pidlFolder, ppidl);
    
    return hr;
}

// Get the target folder for a folder PIDL. This deals with cases where a folder
// is an alias to a real folder, folder shortcuts, the My Documents folder, etc.
STDAPI SHGetTargetFolderPath(LPCITEMIDLIST pidlFolder, LPSTR pszPath)
{
    LPITEMIDLIST pidlTarget;
	
    *pszPath = 0;

    HRESULT hr = SHGetTargetFolderIDList(pidlFolder, &pidlTarget);
    
    if (SUCCEEDED(hr))
    {
        SHGetPathFromIDList(pidlTarget, pszPath);   // Make sure it is a path
        CoTaskMemFree(pidlTarget);
    }
    
    return *pszPath ? S_OK : E_FAIL;
}

//////////////////////////////////////////////////////////////////////////
/////
///// Recopie le backbuffer  l'cran
/////
//////////////////////////////////////////////////////////////////////////

void DBUF_CopyBuffer(TDBufInfo *p_dbi, TBitmap *p_bitmap)
{
	BitBlt(p_dbi->windowcontext, 0, 0, p_bitmap->width, p_bitmap->height, p_dbi->backbufcontext, 0, 0, SRCCOPY);
}

//////////////////////////////////////////////////////////////////////////
/////
///// Alloue les structure ncessaires  un 2ime buffer
/////
//////////////////////////////////////////////////////////////////////////

void DBUF_Init(HWND p_window, TDBufInfo *p_dbi, TBitmap *p_bitmap)
{
	BITMAPINFO BitmapInfo = {0};

	p_dbi->windowcontext = GetDC(p_window);

	// Fill in our BitmapInfo structure (we want a 24 bit image)
	BitmapInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	BitmapInfo.bmiHeader.biCompression = BI_RGB;
	BitmapInfo.bmiHeader.biHeight = -p_bitmap->height;
	BitmapInfo.bmiHeader.biWidth = p_bitmap->width;
	BitmapInfo.bmiHeader.biBitCount = 24;
	BitmapInfo.bmiHeader.biClrUsed = 0;
	BitmapInfo.bmiHeader.biPlanes = 1;

	p_dbi->backbufbitmap = CreateDIBSection(p_dbi->windowcontext, &BitmapInfo, DIB_RGB_COLORS, (void **)&p_bitmap->pMem, 0,0);
	p_dbi->backbufcontext = CreateCompatibleDC(p_dbi->windowcontext);
	p_dbi->oldbitmap = (HBITMAP)SelectObject(p_dbi->backbufcontext, p_dbi->backbufbitmap);
}

//////////////////////////////////////////////////////////////////////////
/////
///// Libre les structures de double buffering
/////
//////////////////////////////////////////////////////////////////////////

void DBUF_Free(HWND p_window, TDBufInfo *p_dbi, TBitmap *p_bitmap)
{
	SelectObject(p_dbi->backbufcontext, p_dbi->oldbitmap);
	DeleteObject(p_dbi->backbufbitmap);
	DeleteDC(p_dbi->backbufcontext);
	ReleaseDC(p_window, p_dbi->windowcontext);
}




//////////////////////////////////////////////////////////////////////////
/////
///// Affiche un requester de fichiers
/////
//////////////////////////////////////////////////////////////////////////

void WIN_FileRequest(HWND p_window, char *p_pResult, u32 p_type, char *p_pTitle, char *p_pFilter, char *p_pInitdir, char *p_pInitfile)
{
	if (p_type != FR_DIR)
	{
		OPENFILENAME open_params;
		int	res;

		memset(&open_params, 0, sizeof(open_params));
		open_params.lStructSize = sizeof(OPENFILENAME);
		open_params.hwndOwner = p_window;
		open_params.lpstrFilter = p_pFilter;
		open_params.nMaxFile = (p_type == FR_MULTI ? MAX_PATH*100 : MAX_PATH); // The size of the "file_name" buffer
		open_params.lpstrInitialDir = p_pInitdir;
		open_params.lpstrFile = p_pResult;

		if (p_pResult != NULL)
		{
			if (p_pInitfile != NULL)
			{
				strcpy(p_pResult, p_pInitfile);
			}
			else
			{
				*p_pResult = '\0';
			}
		}

		open_params.lpstrTitle = p_pTitle;

		// OFN_FILEMUSTEXIST -- Le fichier doit exister
		// OFN_PATHMUSTEXIST -- Le path doit exister
		// OFN_NOCHANGEDIR -- Remet le rpertoire d'origine
		// OFN_HIDEREADONLY -- permet d'ouvrir les fichiers "read only"
		if (p_type == FR_LOAD)
		{
			open_params.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_NOCHANGEDIR | OFN_HIDEREADONLY;
			res = GetOpenFileName(&open_params);
		}
		else if (p_type == FR_SAVE)
		{
			open_params.Flags = OFN_PATHMUSTEXIST | OFN_NOCHANGEDIR | OFN_HIDEREADONLY;
			res = GetSaveFileName(&open_params);
		}
		else
		{
			open_params.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_NOCHANGEDIR | OFN_HIDEREADONLY | OFN_ALLOWMULTISELECT | OFN_EXPLORER;
			res = GetOpenFileName(&open_params);
		}

		if (res)
		{
			strcpy(p_pResult, open_params.lpstrFile);
		}
		else
		{
			*p_pResult = 0;
		}
	}
	else
	{
		BROWSEINFO bi;
		LPITEMIDLIST pid;

		CoInitialize(NULL);
		memset(&bi, 0, sizeof(bi));
		bi.hwndOwner = g_Win.window;
		bi.lpszTitle = p_pTitle;
		bi.ulFlags = BIF_USENEWUI;
		pid = SHBrowseForFolder(&bi);

		if (pid != NULL)
		{
			SHGetPathFromIDList(pid, p_pResult);
		}
		else
		{
			*p_pResult = 0;
		}
	}
}

/*
#define OFN_ALLOWMULTISELECT         0x00000200
*/

//////////////////////////////////////////////////////////////////////////
/////
///// Ouvre une fentre pour laquelle on passe un rectangle de dimension
/////
//////////////////////////////////////////////////////////////////////////

HWND WIN_OpenWindowRect(HINSTANCE p_hInstance, RECT *p_pRect, char *p_pCursor, HBRUSH p_background, u32 p_menu, u32 p_icon, TMenuCallback p_mcb, int p_style, TMsgCB p_mscb)
{
	int	winwidth, winheight;
	u32		i;
	WNDCLASSEX  wndclass = {sizeof (wndclass),				/// size
							CS_HREDRAW | CS_VREDRAW,		/// style
							WIN_StdMessages,				/// procedure
							0,								/// bytes to add to the structure
							0,								/// bytes to add to the structure
							p_hInstance,					/// instance
							(p_icon != 0 ? (HICON)LoadImage(p_hInstance,MAKEINTRESOURCE(p_icon),IMAGE_ICON,0,0,LR_DEFAULTSIZE) : 0),										/// icone
							LoadCursor(NULL, (p_pCursor == NULL ? IDC_ARROW : p_pCursor)),	/// curseur
							(HBRUSH) p_background,	/// background
							(p_menu != NULL ? MAKEINTRESOURCE(p_menu) : NULL),	/// menu
							"Window Class",					/// class name
							(p_icon != 0 ? (HICON)LoadImage(p_hInstance,MAKEINTRESOURCE(p_icon),IMAGE_ICON,0,0,LR_DEFAULTSIZE) : LoadIcon(NULL, IDI_WINLOGO))}; /// small icon

	g_Win.menucallback = p_mcb;
	g_Win.msgcallback = p_mscb;

	// Register the window class with the operating system
	RegisterClassEx (&wndclass);

	winwidth = p_pRect->right - p_pRect->left;
	winheight = p_pRect->bottom - p_pRect->top;

	// Now, we actually create the window
	g_Win.window = CreateWindow ("Window Class",					// g_Win.window class name 
								"",									// g_Win.window's Title    
								p_style,							// g_Win.window style	- This style won't allow the g_Win.window to resize
								p_pRect->left,
								p_pRect->top,
								winwidth,
								winheight,
								NULL,								// This is the parent g_Win.window handle.  
								NULL,								// This is the g_Win.window menu handle
								p_hInstance,						// This is the programs instance handle.
								NULL);								// We don't want to pass any extra data in, so NULL

	// efface les touches
	for (i=0; i<0x100; i++)
	{
		g_Win.key[i] = KEY_NONE;
	}

	// 1/50s timer
	SetTimer(g_Win.window, 0, 20, (TIMERPROC)NULL);
	g_Win.frame = 0;
	return (g_Win.window);
}

//////////////////////////////////////////////////////////////////////////
/////
///// Ouvre une fentre
/////
//////////////////////////////////////////////////////////////////////////

HWND WIN_OpenWindow(HINSTANCE p_hInstance, u16 p_width, u16 p_height, char *p_pCursor, HBRUSH p_background, u32 p_menu, u32 p_icon, TMenuCallback p_mcb)
{
	int	style = WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_BORDER|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_MAXIMIZE;
	RECT	wantedrect;
	int	winposx, winposy, winwidth, winheight;

	/// trouve les dimensions de la fentre en prenant en compte les bordures
	wantedrect.top = 0;
	wantedrect.left = 0;
	wantedrect.right = p_width;
	wantedrect.bottom = p_height;
	AdjustWindowRect(&wantedrect, style, (p_menu != NULL ? TRUE : FALSE));

	winwidth = wantedrect.right - wantedrect.left;
	winheight = wantedrect.bottom - wantedrect.top;

	/// centre la fentre
	winposx = (GetSystemMetrics(SM_CXSCREEN) - winwidth) / 2;
	winposy = (GetSystemMetrics(SM_CYSCREEN) - winheight) / 2;

	wantedrect.top += winposx;
	wantedrect.left += winposy;
	wantedrect.right += winposx;
	wantedrect.bottom += winposy;

	g_Win.window = WIN_OpenWindowRect(p_hInstance, &wantedrect, p_pCursor, p_background, p_menu, p_icon, p_mcb, style, NULL);

	// Init de la bitmap des DC, etc...
	BITMAP_Init(&g_Win.bitmap, p_width, p_height, BITMAP_BGR);
	DBUF_Init(g_Win.window, &g_Win.dbuf, &g_Win.bitmap);

	return (g_Win.window);
}

//////////////////////////////////////////////////////////////////////////
/////
///// MyPrint()
///// Affiche un texte dans une fentre
/////
//////////////////////////////////////////////////////////////////////////
void WIN_Print(u32 p_x, u32 p_y, const char *p_pFormat, ...)
{
	va_list arglist;
	static char	pMystring[256];

	va_start(arglist, p_pFormat);
	vsprintf(pMystring, p_pFormat, arglist);
	va_end(arglist);
	TextOut(g_Win.dbuf.backbufcontext, p_x, p_y, pMystring, (int)strlen(pMystring));
}

//////////////////////////////////////////////////////////////////////////
/////
///// WIN_WaitForFrame()
///// Gestion  chaque trame
///// Renvoie TRUE si il faut quitter
/////
//////////////////////////////////////////////////////////////////////////
u32 WIN_WaitForFrame(u32 p_isRedraw)
{
	MSG		msg;
	u32		i;

	// mise  jour des touches
	for (i=0; i<0x100; i++)
	{
		if (g_Win.key[i] == KEY_TRIG)
		{
			g_Win.key[i] = KEY_CONT;
		}
		else if (g_Win.key[i] == KEY_RELEASE)
		{
			g_Win.key[i] = KEY_NONE;
		}
	}

	WaitMessage();

	// traitement des messages
	g_Win.timer = FALSE;

	while (1)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))	// Check if there is a window message and remove it from the queue
		{
			if (msg.message == WM_QUIT)
			{
				return (TRUE);
			}

			TranslateMessage (&msg);	// This Translates messages so windows understands them.
			DispatchMessage (&msg);		// This sends the messages to WIN_StdMessages().
		}
		else
		{
			if (g_Win.timer == TRUE)
			{
				if (p_isRedraw == TRUE)
				{
					WIN_Redraw();
				}
				g_Win.frame++;
				return (FALSE);
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// gestion des messages de la fentre
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK WIN_StdMessages(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	// Check which message was passed in
	switch (iMsg)
	{
		// This message is sent when the window is created (CreateWindow())
		case WM_CREATE:
			break;

		case WM_KEYDOWN:
			if (wParam < 0x100)
			{
				if (g_Win.key[wParam] == KEY_NONE)
				{
					g_Win.key[wParam] = KEY_TRIG;
				}
				else
				{
					g_Win.key[wParam] = KEY_REPEAT;
				}
			}
			break;

		case WM_KEYUP:
			if (wParam < 0x100)
			{
				g_Win.key[wParam] = KEY_RELEASE;
			}
			break;

		// This message is sent when the window is destroyed
		case WM_DESTROY:
			WIN_StdClose(hwnd);
			break;

		case WM_PAINT:
		case WM_ACTIVATEAPP:
//		case WM_MOVE:
			WIN_Redraw();
			break;

		case WM_TIMER:
			if (wParam == 0)
			{
				g_Win.timer = TRUE;
			}
			break;

		case WM_COMMAND:									// This message is sent when the user selects a command item from a menu, and some other things.
			if (g_Win.menucallback != NULL)
			{
				g_Win.menucallback(hwnd, LOWORD(wParam));
			}
			break;

		case WM_HSCROLL:
		case WM_VSCROLL:
			if (g_Win.msgcallback != NULL)
			{
				g_Win.msgcallback(hwnd, iMsg, wParam, lParam);
			}
			break;
	}

	// Process/handle the default messages
	return DefWindowProc (hwnd, iMsg, wParam, lParam);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// WIN_Close
/// fermeture de la fentre
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void WIN_StdClose(HWND p_hwnd)
{
	DBUF_Free(p_hwnd, &g_Win.dbuf, &g_Win.bitmap);
	ChangeDisplaySettings(NULL,0);

	// Post the QUIT message to the window (0 = WM_QUIT)															
	PostQuitMessage(0);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// WIN_Redraw
/// mise  jour de la fentre
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void WIN_Redraw(void)
{
	DBUF_CopyBuffer(&g_Win.dbuf, &g_Win.bitmap);
}
