// retouche.cpp : Defines the entry point for the console application.
//

//  faire: indexs (dans tga aussi !)

#include "stdafx.h"
#include <math.h>
#include <stdlib.h>
#include "mem.h"
#include "bitmap.h"
#include "tga.h"
#include "tiff.h"

typedef struct
{
	u8	r, g, b;
} TColor;

struct SParams
{
	int		detect_width;
	int		detect_diff;
	int		blur_radius;
	int		blur_limit;
	int		vision_limit1;
	int		vision_limit2;
	int		vision_width1;
	int		vision_width2;
	int		vision_width3;
	double	vision_angle;
	char	repertoire1[256];
	char	repertoire2[256];
};

struct SParams	p;
double	pi = 3.14159265359;

/******************************************************************************
 *
 * Recopie une chaine de caractres en enlevant les espaces  la fin et les guillemets
 *
 ******************************************************************************/
void GetRep(char *p_pSrc, char *p_pDst)
{
	if (*p_pSrc == '\"')
	{
		p_pSrc++;
	}

	strcpy(p_pDst, p_pSrc);

	while (p_pDst[strlen(p_pDst)-1] == ' ')
	{
		p_pDst[strlen(p_pDst)-1] = 0;
	}

	if (p_pDst[strlen(p_pDst)-1] == '\"')
	{
		p_pDst[strlen(p_pDst)-1] = 0;
	}
}

/******************************************************************************
 *
 * Lit les paramtres dans le fichier de config
 *
 ******************************************************************************/
void GetParams(void)
{
	FILE	*pFile;
	char	pTmp[256];
	char	pTmp2[256];
	char	*pPtr;

	p.detect_width = 20;
	p.detect_diff = 25;
	p.vision_limit1 = 64;
	p.vision_limit2 = 3;
	p.vision_width1 = 16;
	p.vision_width2 = 16;
	p.vision_width3 = 128;
	p.vision_angle = 3.0;
	p.blur_radius = 3;
	p.blur_limit = 50;
	strcpy(p.repertoire1, "c:/t/");
	strcpy(p.repertoire2, "c:/trot/");

	pFile = fopen("config.txt", "r");

	if (pFile == NULL)
	{
		return;
	}

	while (fgets(pTmp, 256, pFile) != NULL)
	{
		while(pTmp[strlen(pTmp)-1] == 0xA || pTmp[strlen(pTmp)-1] ==0xD)
		{
			pTmp[strlen(pTmp)-1] = 0;
		}

		pPtr = strchr(pTmp, '=');

		if (pPtr != NULL)
		{
			memcpy(pTmp2, pTmp, pPtr-pTmp);
			pTmp2[pPtr-pTmp] = 0;
			strupr(pTmp2);

			if (strcmp(pTmp2, "DETECT_WIDTH") == 0)
			{
				sscanf(pPtr+1, "%d", &p.detect_width);
			}
			else if (strcmp(pTmp2, "DETECT_DIFF") == 0)
			{
				sscanf(pPtr+1, "%d", &p.detect_diff);
			}
			else if (strcmp(pTmp2, "VISION_LIMIT1") == 0)
			{
				sscanf(pPtr+1, "%d", &p.vision_limit1);
			}
			else if (strcmp(pTmp2, "VISION_LIMIT2") == 0)
			{
				sscanf(pPtr+1, "%d", &p.vision_limit2);
			}
			else if (strcmp(pTmp2, "VISION_WIDTH1") == 0)
			{
				sscanf(pPtr+1, "%d", &p.vision_width1);
			}
			else if (strcmp(pTmp2, "VISION_WIDTH2") == 0)
			{
				sscanf(pPtr+1, "%d", &p.vision_width2);
			}
			else if (strcmp(pTmp2, "VISION_WIDTH3") == 0)
			{
				sscanf(pPtr+1, "%d", &p.vision_width3);
			}
			else if (strcmp(pTmp2, "VISION_ANGLE") == 0)
			{
				float tmp = 0.0;

				sscanf(pPtr+1, "%f", &tmp);
				p.vision_angle = tmp;
			}
			else if (strcmp(pTmp2, "BLUR_RADIUS") == 0)
			{
				sscanf(pPtr+1, "%d", &p.blur_radius);
			}
			else if (strcmp(pTmp2, "BLUR_LIMIT") == 0)
			{
				sscanf(pPtr+1, "%d", &p.blur_limit);
			}
			else if (strcmp(pTmp2, "REPERTOIRE1") == 0)
			{
				pPtr++;
				GetRep(pPtr, p.repertoire1);
			}
			else if (strcmp(pTmp2, "REPERTOIRE2") == 0)
			{
				pPtr++;
				GetRep(pPtr, p.repertoire2);
			}
			else
			{
				printf("paramtre inconnu \"%s\"\n", pTmp2);
			}
		}
	}
	fclose(pFile);
}

/******************************************************************************
 *
 * Lit un pixel d'une bitmap
 *
 ******************************************************************************/
void GetPx(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	switch (p_pBm->format)
	{
		case BITMAP_GREY:
			pAddr = BITMAP_ADDR8(p_pBm, p_x, p_y);
			p_pCol->r = *pAddr;
			p_pCol->g = *pAddr;
			p_pCol->b = *pAddr;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
			p_pCol->r = *pAddr++;
			p_pCol->g = *pAddr++;
			p_pCol->b = *pAddr;
			break;
	}
}

/******************************************************************************
 *
 * Ecrit un pixel dans une bitmap
 *
 ******************************************************************************/
void SetPx(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	switch (p_pBm->format)
	{
		case BITMAP_GREY:
			*BITMAP_ADDR8(p_pBm, p_x, p_y) = p_pCol->r;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
			*pAddr++ = p_pCol->r;
			*pAddr++ = p_pCol->g;
			*pAddr   = p_pCol->b;
			break;
	}
}

/******************************************************************************
 *
 * Trouve la dernire colonne de la page
 *
 ******************************************************************************/
s32 RecadreDroite(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (x=p_pBm->width-1; x>=0; x--)
	{
		for (y=0; y<p_pBm->height-p.detect_width; y++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx(p_pBm, x, y+i, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (y != p_pBm->height-p.detect_width)
		{
			break;
		}
	}
	return (x+1);
}

/******************************************************************************
 *
 * Trouve la dernire ligne de la page
 *
 ******************************************************************************/
s32 RecadreBas(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (y=p_pBm->height-1; y>=0; y--)
	{
		for (x=0; x<p_pBm->width-p.detect_width; x++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx(p_pBm, x+i, y, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (x != p_pBm->width-p.detect_width)
		{
			break;
		}
	}
	return (y+1);
}

/******************************************************************************
 *
 * Trouve la couleur de fond de l'image (la couleur de la colonne de droite)
 *
 ******************************************************************************/
void GetBGCol(TBitmap *p_pBm, TColor *p_pBcol)
{
	TColor col;
	s32	i, r, g, b;

	r = 0;
	g = 0;
	b = 0;

	for (i=0; i<p_pBm->height; i++)
	{
		GetPx(p_pBm, p_pBm->width-1, i, &col);
		r += col.r;
		g += col.g;
		b += col.b;
	}
	p_pBcol->r = (u8)(r / p_pBm->height);
	p_pBcol->g = (u8)(g / p_pBm->height);
	p_pBcol->b = (u8)(b / p_pBm->height);
}

/******************************************************************************
 *
 * Initialise le noyau du flou gaussien
 *
 ******************************************************************************/
s32 *Gaussian_Init(void)
{
	s32	*pBlur = (s32 *)MEM_Alloc(sizeof(int)*(2*p.blur_radius+1)*(2*p.blur_radius+1));
	s32	*pBlur2 = pBlur;
	double sigma = (double)p.blur_radius / 3.0;
	double e;
	int	x, y;

	if (pBlur != NULL)
	{
		for (y=-p.blur_radius; y<p.blur_radius+1; y++)
		{
			for (x=-p.blur_radius; x<p.blur_radius+1; x++)
			{
				e = (double)(x*x + y*y);
				e = exp(-e / (2.0*sigma*sigma));
				*pBlur2++ = (int)(e * 1024.0);
			}
		}
	}

	return (pBlur);
}

/******************************************************************************
 *
 * Applique le flou gaussien
 *
 ******************************************************************************/
void Gaussian_Process(TBitmap *p_pSrc, TBitmap *p_pDst, s32 *p_pBlur)
{
	s32	x, y, x2, y2;
	s32	r, g, b, k;
	s32	dr, dg, db;
	s32	*pBlur2;
	TColor	colm;
	TColor	col;

	for (y=0; y<p_pDst->height; y++)
	{
		for (x=0; x<p_pDst->width; x++)
		{
			GetPx(p_pSrc, x, y, &colm);
			r = 0;
			g = 0;
			b = 0;
			k = 0;
			pBlur2 = p_pBlur;

			for (y2=-p.blur_radius; y2<p.blur_radius+1; y2++)
			{
				for (x2=-p.blur_radius; x2<p.blur_radius+1; x2++)
				{
					if (x+x2 >= 0 && x+x2 < p_pDst->width &&
						y+y2 >= 0 && y+y2 < p_pDst->height)
					{
						GetPx(p_pSrc, x+x2, y+y2, &col);
						dr = (s32)colm.r - (s32)col.r;
						dg = (s32)colm.g - (s32)col.g;
						db = (s32)colm.b - (s32)col.b;

						if (dr*dr + dg*dg + db*db < p.blur_limit*p.blur_limit)
						{
							r += *pBlur2 * (s32)col.r;
							g += *pBlur2 * (s32)col.g;
							b += *pBlur2 * (s32)col.b;
							k += *pBlur2;
						}
					}
					pBlur2++;
				}
			}
			col.r = (u8)(r / k);
			col.g = (u8)(g / k);
			col.b = (u8)(b / k);
			SetPx(p_pDst, x, y, &col);
		}
	}
}

/******************************************************************************
 *
 * Filtre 1: ne garde que les transitions verticales franches
 *
 ******************************************************************************/
void Filtre1(TBitmap *p_pBm)
{
	s32	x, y, c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			if (y == p_pBm->height-1)
			{
				*pPtr++ = 0;
			}
			else
			{
				c = abs((s32)pPtr[0] - (s32)pPtr[p_pBm->width]);
				*pPtr++ = (c < p.vision_limit1 ? 0 : 255);
			}
		}
	}
}

/******************************************************************************
 *
 * Filtre 2: ne garde que les transition frquentes
 *
 ******************************************************************************/
void Filtre2(TBitmap *p_pBm)
{
	s32	x, y, x2, x3, c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			c = 0;

			if (x < p_pBm->width-p.vision_width1)
			{
				for (x2=0; x2<p.vision_width1-p.vision_limit2; x2++)
				{
					for (x3=0; x3<p.vision_limit2; x3++)
					{
						if (pPtr[x2+x3] == 0)
						{
							break;
						}
					}

					if (x3 == p.vision_limit2)
					{
						c++;
					}
				}
			}
			*pPtr++ = (c != 0 ? 255 : 0);
		}
	}
}

/******************************************************************************
 *
 * Filtre 3: tend les lignes
 *
 ******************************************************************************/
void Filtre3(TBitmap *p_pBm)
{
	s32	x, y, x2, x3;
	u8	*pPtr;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			pPtr = BITMAP_ADDR8(p_pBm, x, y);

			if (pPtr[0] == 0 && x != 0)
			{
				if (pPtr[-1] == 255)
				{
					for (x2=0; x2<p.vision_width2; x2++)
					{
						if (x+x2 >= p_pBm->width)
						{
							x = p_pBm->width;
							break;
						}

						if (pPtr[x2] == 255 || (y != 0 && pPtr[x2-p_pBm->width] == 255) || (y != p_pBm->height-1 && pPtr[x2+p_pBm->width] == 255))
						{
							for (x3=0; x3<x2; x3++)
							{
								*pPtr++ = 1;
								x++;
							}
							pPtr++;
							x++;
							break;
						}
					}
				}
			}
		}
	}
}

/******************************************************************************
 *
 * Filtre 4: lignes d'1 pixel
 *
 ******************************************************************************/
void Filtre4(TBitmap *p_pBm)
{
	s32	x, y;
	u8	c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			c = 0;

			if (y < p_pBm->height-1)
			{
				if (pPtr[0] != 0 && pPtr[p_pBm->width] == 0)
				{
					c = 255;
				}
			}
			*pPtr++ = c;
		}
	}
}

/******************************************************************************
 *
 * Filtre 5: moyenne des angles des lignes les plus longues
 *
 ******************************************************************************/
double Filtre5(TBitmap *p_pBm)
{
	s32	x, y, x2, y2;
	double angle = 0.0;
	double angle2, dist;
	double fac = 0.0;

	for (x=0; x<p_pBm->width; x++)
	{
		for (y=0; y<p_pBm->height; y++)
		{
			if (*BITMAP_ADDR8(p_pBm, x, y) == 255)
			{
				x2 = x;
				y2 = y;

				while (1)
				{
					*BITMAP_ADDR8(p_pBm, x2, y2) = 0;

					if (x2 >= p_pBm->width-1)
					{
						break;
					}

					if (*BITMAP_ADDR8(p_pBm, x2+1, y2) != 0)
					{
						x2++;
					}
					else if (y2 > 0 && *BITMAP_ADDR8(p_pBm, x2+1, y2-1) != 0)
					{
						x2++;
						y2--;
					}
					else if (y2 < p_pBm->height-1 && *BITMAP_ADDR8(p_pBm, x2+1, y2+1) != 0)
					{
						x2++;
						y2++;
					}
					else
					{
						break;
					}
				}

				if (x2 - x >= p.vision_width3)
				{
					angle2 = atan2((double)(y-y2), (double)(x2-x));
					angle2 = angle2 * 180.0 / pi;

					if (fabs(angle2) < p.vision_angle)
					{
						dist = (double)(x2-x);
						angle += dist * angle2;
						fac += dist;
					}
				}
			}
		}
	}

	if (fac != 0.0)
	{
		angle = angle / fac;
	}

	return (angle);
}

/******************************************************************************
 *
 * calcule un coefficient pour l'interpolation bicubique en fonction de la distance du point
 *
 ******************************************************************************/
double Rotate_CalcCoef(double p_dist)
{
	double	dist2 = p_dist * p_dist;

	if (p_dist < 1.0)
	{
		return (1.5 * dist2*p_dist - 2.5 * dist2 + 1.0);
	}
	else if (p_dist < 2.0)
	{
		return (-0.5 * dist2*p_dist + 2.5 * dist2 - 4.0 * p_dist + 2.0);
	}
	else
	{
		return (0.0);
	}
}

/******************************************************************************
 *
 * limite une valeur entre 0 et 255
 *
 ******************************************************************************/
u8 limit(s32 p_val)
{
	if (p_val < 0)
	{
		return (0);
	}
	else if (p_val > 255)
	{
		return (255);
	}
	else
	{
		return ((u8)p_val);
	}
}

/******************************************************************************
 *
 * trouve l'image d'un point aprs la rotation, en utilisant une interpolation bicubique
 *
 ******************************************************************************/
void RotatePoint(TBitmap *p_pSrc, double p_x, double p_y, TColor *p_pBcol, TColor *p_pCol)
{
	s32	intx = (int)floor(p_x);
	s32	inty = (int)floor(p_y);
	s32	srcx, srcy;
	double	dist, coef;
	double	fr, fg, fb, fac;
	TColor	tmpc;

	fac = 0.0;
	fr = 0.0;
	fg = 0.0;
	fb = 0.0;

	for (srcx=intx-1; srcx<intx+3; srcx++)
	{
		for (srcy=inty-1; srcy<inty+3; srcy++)
		{
			dist = fabs((double)srcx + 0.5 - p_x);
			coef = Rotate_CalcCoef(dist);
			dist = fabs((double)srcy + 0.5 - p_y);
			coef *= Rotate_CalcCoef(dist);
			fac += coef;

			if (srcx >= 0 && srcx < p_pSrc->width &&
				srcy >= 0 && srcy < p_pSrc->height)
			{
				GetPx(p_pSrc, srcx, srcy, &tmpc);
				fr += coef * (double)tmpc.r;
				fg += coef * (double)tmpc.g;
				fb += coef * (double)tmpc.b;
			}
			else
			{
				fr += coef * (double)p_pBcol->r;
				fg += coef * (double)p_pBcol->g;
				fb += coef * (double)p_pBcol->b;
			}
		}
	}
	p_pCol->r = limit((s32)floor(fr / fac));
	p_pCol->g = limit((s32)floor(fg / fac));
	p_pCol->b = limit((s32)floor(fb / fac));
}

/******************************************************************************
 *
 * Rotation de l'image
 *
 ******************************************************************************/
void Rotate(TBitmap *p_pSrc, TBitmap *p_pDst, double p_angle, TColor *p_pBcol)
{
	double	center_x, center_y;
	double	src_x, src_y;
	double	savex, savey;
	double	incx, incy;
	double	angle, angle2;
	double	dist;
	s32	dst_x, dst_y;
	TColor	col;

	angle = p_angle * pi / 180.0;
	incx = cos(angle);
	incy = -sin(angle);

	center_x = (float)p_pSrc->width / 2.0 - 0.5;
	center_y = (float)p_pSrc->height / 2.0 - 0.5;
	dist = sqrt(center_x*center_x + center_y*center_y);
	angle2 = atan2(-center_y, center_x);
	savex = center_x+0.5 - dist * cos(angle2+angle);
	savey = center_y+0.5 + dist * sin(angle2+angle);

	for (dst_y=0; dst_y<p_pDst->height; dst_y++)
	{
		src_x = savex;
		src_y = savey;
		savex -= incy;
		savey += incx;

		for (dst_x=0; dst_x<p_pDst->width; dst_x++)
		{
			RotatePoint(p_pSrc, src_x, src_y, p_pBcol, &col);
			SetPx(p_pDst, dst_x, dst_y, &col);
			src_x += incx;
			src_y += incy;
		}
	}
}

/******************************************************************************
 *
 * Point d'entre du programme
 *
 ******************************************************************************/
int _tmain(int argc, _TCHAR* argv[])
{
	s32		newheight, newwidth;
	s32		*pBlur;
	u32		isFileLoaded;
	char	pFilename[256];
	char	*pTmp;
	double	angle;
	TBitmap	src_bitmap;
	TBitmap	work_bitmap;
	TBitmap	work2_bitmap;
	TColor	back_col;

	// init
	if (argc != 2)
	{
		printf("retouche v2.0 par Frederic Goset\n");
		printf("  usage: retouche <nom de fichier>\n");
		return (0);
	}

	GetParams();

	// chargement
	isFileLoaded = TIFF_Load(argv[1], &src_bitmap);

	if (isFileLoaded == FALSE)
	{
		isFileLoaded = TGA_Load(argv[1], &src_bitmap);
	}

	// conversion du format
	if (src_bitmap.format == BITMAP_RGB)
	{
		// la routine de sauvegarde TGA n'accepte pas le format RGB: on le convertit en BGR
		memcpy(&work_bitmap, &src_bitmap, sizeof(TBitmap));
		src_bitmap.pMem = NULL;
		src_bitmap.pPalette = NULL;
		BITMAP_Alloc(&src_bitmap, work_bitmap.width, work_bitmap.height, BITMAP_BGR);
		BITMAP_Blt(&work_bitmap, &src_bitmap, 0, 0);
		BITMAP_Free(&work_bitmap);
	}
	else if (src_bitmap.format == BITMAP_256)
	{
		// le programme n'est pas conu pour les images en 256 couleurs: on les convertit en tons de gris
		memcpy(&work_bitmap, &src_bitmap, sizeof(TBitmap));
		src_bitmap.pMem = NULL;
		src_bitmap.pPalette = NULL;
		BITMAP_Alloc(&src_bitmap, work_bitmap.width, work_bitmap.height, BITMAP_GREY);
		BITMAP_Blt(&work_bitmap, &src_bitmap, 0, 0);
		BITMAP_Free(&work_bitmap);
	}

	if (isFileLoaded == TRUE)
	{
		// get background color
		GetBGCol(&src_bitmap, &back_col);

		// recadrage
		newwidth = RecadreDroite(&src_bitmap, &back_col);
		newheight = RecadreBas(&src_bitmap, &back_col);

		// init du flou
		pBlur = Gaussian_Init();

		if (pBlur != NULL)
		{
			// application du flou
			if (BITMAP_Alloc(&work_bitmap, (u16)newwidth, (u16)newheight, src_bitmap.format) == TRUE)
			{
				Gaussian_Process(&src_bitmap, &work_bitmap, pBlur);
				BITMAP_Free(&src_bitmap);
				MEM_Free((void **)&pBlur);

				// sauvegarde 1
				strcpy(pFilename, p.repertoire1);
				strcat(pFilename, argv[1]);
				pTmp = strrchr(pFilename, '.');
				strcpy(pTmp, ".tga");
				TGA_Save(pFilename, &work_bitmap);

				// cre le buffer de travail pour la dtection de l'angle
				if (BITMAP_Alloc(&work2_bitmap, (u16)newwidth, (u16)newheight, BITMAP_GREY) == TRUE)
				{
					BITMAP_Blt(&work_bitmap, &work2_bitmap, 0, 0);

					// ne garde que les transitions verticales franches
					Filtre1(&work2_bitmap);

					// ne garde que les transition frquentes
					Filtre2(&work2_bitmap);

					// tend les lignes
					Filtre3(&work2_bitmap);

					// lignes d'1 pixel
					Filtre4(&work2_bitmap);

					// moyenne des angles des lignes les plus longues
					angle = Filtre5(&work2_bitmap);
					printf("%f\n", angle);
					BITMAP_Free(&work2_bitmap);

					// rotation
					if (BITMAP_Alloc(&work2_bitmap, (u16)newwidth, (u16)newheight, work_bitmap.format) == TRUE)
					{
						BITMAP_Clr(&work2_bitmap, back_col.r, back_col.g, back_col.b);
						Rotate(&work_bitmap, &work2_bitmap, angle, &back_col);
						BITMAP_Free(&work_bitmap);

						// recadrage 2
						newwidth = RecadreDroite(&work2_bitmap, &back_col);
						newheight = RecadreBas(&work2_bitmap, &back_col);

						// recopie l'image recadre
						if (BITMAP_Alloc(&work_bitmap, (u16)newwidth, (u16)newheight, work2_bitmap.format) == TRUE)
						{
							BITMAP_Blt(&work2_bitmap, &work_bitmap, 0, 0);
							BITMAP_Free(&work2_bitmap);

							// sauvegarde 2
							strcpy(pFilename, p.repertoire2);
							strcat(pFilename, argv[1]);
							pTmp = strrchr(pFilename, '.');
							strcpy(pTmp, ".tga");
							TGA_Save(pFilename, &work_bitmap);

							printf("%s done.\n", argv[1]);

							BITMAP_Free(&work_bitmap);
						}
						BITMAP_Free(&work2_bitmap);
					}
					BITMAP_Free(&work2_bitmap);
				}
				BITMAP_Free(&work_bitmap);
			}
			MEM_Free((void **)&pBlur);
		}
		BITMAP_Free(&src_bitmap);
	}
	return 0;
}
