#ifndef	TIFF_H
#define	TIFF_H

#include "MyTypes.h"
#include "bitmap.h"

// data types of entries
enum
{
	TIFF_ENTRYTYPE_NULL=0,
	TIFF_ENTRYTYPE_BYTE,
	TIFF_ENTRYTYPE_ASCII,
	TIFF_ENTRYTYPE_SHORT,
	TIFF_ENTRYTYPE_LONG,
	TIFF_ENTRYTYPE_RATIONAL,
	TIFF_ENTRYTYPE_SBYTE,
	TIFF_ENTRYTYPE_UNDEFINED,
	TIFF_ENTRYTYPE_SSHORT,
	TIFF_ENTRYTYPE_SLONG,
	TIFF_ENTRYTYPE_SRATIONAL,
	TIFF_ENTRYTYPE_FLOAT,
	TIFF_ENTRYTYPE_DOUBLE,
	TIFF_ENTRYTYPE_INVALID
};

// endian types
#define	TIFF_BIG_ENDIAN		0
#define	TIFF_LITTLE_ENDIAN	1

// entries tags
#define	TIFF_TAG_NEWSUBFILETYPE					0x00FE
#define	TIFF_TAG_SUBFILETYPE					0x00FF
#define	TIFF_TAG_IMAGEWIDTH						0x0100
#define	TIFF_TAG_IMAGEHEIGHT					0x0101
#define	TIFF_TAG_BITSPERSAMPLE					0x0102
#define	TIFF_TAG_COMPRESSION					0x0103
#define	TIFF_TAG_PHOTOMETRICINTERPRETATION		0x0106
#define	TIFF_TAG_THRESHHOLDING					0x0107
#define	TIFF_TAG_CELLWIDTH						0x0108
#define	TIFF_TAG_CELLLENGTH						0x0109
#define	TIFF_TAG_FILLORDER						0x010A
#define	TIFF_TAG_IMAGEDESCRIPTION				0x010E
#define	TIFF_TAG_MAKE							0x010F
#define	TIFF_TAG_MODEL							0x0110
#define	TIFF_TAG_STRIPOFFSETS					0x0111
#define	TIFF_TAG_ORIENTATION					0x0112
#define	TIFF_TAG_SAMPLESPERPIXEL				0x0115
#define	TIFF_TAG_ROWSPERSTRIP					0x0116
#define	TIFF_TAG_STRIPBYTECOUNTS				0x0117
#define	TIFF_TAG_MINSAMPLEVALUE					0x0118
#define	TIFF_TAG_MAXSAMPLEVALUE					0x0119
#define	TIFF_TAG_XRESOLUTION					0x011A
#define	TIFF_TAG_YRESOLUTION					0x011B
#define	TIFF_TAG_PLANARCONFIGURATION			0x011C
#define	TIFF_TAG_FREEOFFSETS					0x0120
#define	TIFF_TAG_FREEBYTECOUNTS					0x0121
#define	TIFF_TAG_GRAYRESPONSEUNIT				0x0122
#define	TIFF_TAG_GRAYRESPONSECURVE				0x0123
#define	TIFF_TAG_RESOLUTIONUNIT					0x0128
#define	TIFF_TAG_TRANSFERFUNCTION				0x012D
#define	TIFF_TAG_SOFTWARE						0x0131
#define	TIFF_TAG_DATETIME						0x0132
#define	TIFF_TAG_ARTIST							0x013B
#define	TIFF_TAG_HOSTCOMPUTER					0x013C
#define	TIFF_TAG_PREDICTOR						0x013D
#define	TIFF_TAG_WHITEPOINT						0x013E
#define	TIFF_TAG_PRIMARYCHROMATICITIES			0x013F
#define	TIFF_TAG_COLORMAP						0x0140
#define	TIFF_TAG_HALFTONEHINTS					0x0141
#define	TIFF_TAG_TILEWIDTH						0x0142
#define	TIFF_TAG_TILELENGTH						0x0143
#define	TIFF_TAG_TILEOFFSETS					0x0144
#define	TIFF_TAG_TILEBYTECOUNTS					0x0145
#define	TIFF_TAG_INKSET							0x014C
#define	TIFF_TAG_INKNAMES						0x014D
#define	TIFF_TAG_NUMBEROFINKS					0x014E
#define	TIFF_TAG_DOTRANGE						0x0150
#define	TIFF_TAG_TARGETPRINTER					0x0151
#define	TIFF_TAG_EXTRASAMPLES					0x0152
#define	TIFF_TAG_SAMPLEFORMAT					0x0153
#define	TIFF_TAG_SMINSAMPLEVALUE				0x0154
#define	TIFF_TAG_SMAXSAMPLEVALUE				0x0155
#define	TIFF_TAG_TRANSFERRANGE					0x0156
#define	TIFF_TAG_JPEGPROC						0x0200
#define	TIFF_TAG_JPEGINTERCHANGEFORMAT			0x0201
#define	TIFF_TAG_JPEGINTERCHANGEFORMATLENGTH	0x0202
#define	TIFF_TAG_JPEGRESTARTINTERVAL			0x0203
#define	TIFF_TAG_JPEGLOSSLESSPREDICTORS			0x0205
#define	TIFF_TAG_JPEGPOINTTRANSFORMS			0x0206
#define	TIFF_TAG_JPEGQTABLES					0x0207
#define	TIFF_TAG_JPEGDCTABLES					0x0208
#define	TIFF_TAG_JPEGACTABLES					0x0209
#define	TIFF_TAG_YCBCRCOEFFICIENTS				0x0211
#define	TIFF_TAG_YCBCRSUBSAMPLING				0x0212
#define	TIFF_TAG_YCBCRPOSITIONING				0x0213
#define	TIFF_TAG_REFERENCEBLACKWHITE			0x0214
#define	TIFF_TAG_COPYRIGHT						0x8298

// compression types
#define	TIFF_COMPRESSION_NONE		1
#define	TIFF_COMPRESSION_HUFFMAN	2
#define	TIFF_COMPRESSION_FAX3		3
#define	TIFF_COMPRESSION_FAX4		4
#define	TIFF_COMPRESSION_LZW		5
#define	TIFF_COMPRESSION_JPEG		6
#define	TIFF_COMPRESSION_PACKBITS	32773

// photometric interpretation
#define	TIFF_PHOTO_INVERTGRAY		0	/// White Is Zero
#define	TIFF_PHOTO_GRAY				1	/// Black Is Zero
#define	TIFF_PHOTO_RGB				2
#define	TIFF_PHOTO_COLORMAP			3
#define	TIFF_PHOTO_SEPARATED		5
#define	TIFF_PHOTO_YCBCR			6
#define	TIFF_PHOTO_CIELAB			8

// LZW compression codes
#define	TIFF_LZW_CLEARCODE			256
#define	TIFF_LZW_EOICODE			257
#define	TIFF_LZW_FIRST_MULTI		258

// fill orders
#define	TIFF_ORDER_HIGHBITSFIRST	1
#define	TIFF_ORDER_LOWBITSFIRST		2

// predictors
#define	TIFF_PREDICTOR_NONE			1
#define	TIFF_PREDICTOR_DIFFERENCE	2

// huffman colors
#define	TIFF_HUFF_WHITE		0
#define	TIFF_HUFF_BLACK		1

#define	TIFF_HUFF_MAXCODELEN	15

// donnes pour la dcompression LZW
typedef struct STIFF_LZW
{
	u16	pCodes[4096];
	u8	pChars[4096];
	u32	src_off;
	u32	dst_off;
	u16	table_size;
	u8	src_bit;
	u8	code_size;
} TTIFF_LZW;

// infos pour un strip
typedef struct STiffStrip
{
	u32		offset;
	u32		bytes;
} TTiffStrip;

// toutes les infos du fichier
typedef struct STiffInfo
{
	char		pFilename[256];
	u8			pPal[256][3];
	FILE		*pFile;
	TBitmap		*pBm;
	TTiffStrip	*pStrips;
	u32			rows_per_strip;
	u32			qt_strips;
	u32			bytes_per_row;
	u16			compression;
	u8			bits_per_sample;
	u8			fill_order;
	u8			samples_per_pixel;
	u8			predictor;
	u8			photo_interpret;
	u8			endian:1;
} TTiffInfo;

u32 TIFF_Load(char *p_pFilename, TBitmap *p_pBitmap);

#endif
