// retouche.cpp : Defines the entry point for the console application.
//

//  faire: indexs (dans tga aussi !)

#include "stdafx.h"
#include <math.h>
#include <stdlib.h>
#include "mem.h"
#include "bitmap.h"
#include "error.h"
#include "tga.h"
#include "tiff.h"

typedef struct
{
	u8	r, g, b;
} TColor;

#define	BIT_HAUT	1
#define	BIT_BAS		2
#define	BIT_GAUCHE	4
#define	BIT_DROITE	8

struct SParams
{
	int		detect_width;
	int		detect_diff;
	int		blur_radius;
	int		blur_limit;
	int		vision_limit1;
	int		vision_limit2;
	int		vision_width1;
	int		vision_width2;
	int		vision_width3;
	u8		cote_recadrage;
	u8		cote_fond;
	u8		isRecadrage1;
	u8		isFlou;
	u8		isSauvegarde1;
	u8		isRotation;
	u8		isRecadrage2;
	u8		isSauvegarde2;
	u8		isShowFilters;
	double	vision_angle;
	char	repertoire1[256];
	char	repertoire2[256];
	char	file[256];
};

struct SParams	p;
double	pi = 3.14159265359;

/******************************************************************************
 *
 * Recopie une chaine de caractres en enlevant les espaces  la fin et les guillemets
 *
 ******************************************************************************/
void GetRep(char *p_pSrc, char *p_pDst)
{
	if (*p_pSrc == '\"')
	{
		p_pSrc++;
	}

	strcpy(p_pDst, p_pSrc);

	while (p_pDst[strlen(p_pDst)-1] == ' ')
	{
		p_pDst[strlen(p_pDst)-1] = 0;
	}

	if (p_pDst[strlen(p_pDst)-1] == '\"')
	{
		p_pDst[strlen(p_pDst)-1] = 0;
	}
}

/******************************************************************************
 *
 * Lit les paramtres dans le fichier de config
 *
 ******************************************************************************/
enum
{
	TYPE_INT=0,
	TYPE_FLOAT,
	TYPE_STRING,
	TYPE_COTE,
	TYPE_BINAIRE
};

typedef struct SPrm
{
	char	pCmd[40];
	int		type;
	void	*pVar;
} TPrm;

static TPrm g_pCmd[] =
{
	{"DETECT_WIDTH",	TYPE_INT,		&p.detect_width},
	{"DETECT_DIFF",		TYPE_INT,		&p.detect_diff},
	{"VISION_LIMIT1",	TYPE_INT,		&p.vision_limit1},
	{"VISION_LIMIT2",	TYPE_INT,		&p.vision_limit2},
	{"VISION_WIDTH1",	TYPE_INT,		&p.vision_width1},
	{"VISION_WIDTH2",	TYPE_INT,		&p.vision_width2},
	{"VISION_WIDTH3",	TYPE_INT,		&p.vision_width3},
	{"VISION_ANGLE",	TYPE_FLOAT,		&p.vision_angle},
	{"BLUR_RADIUS",		TYPE_INT,		&p.blur_radius},
	{"BLUR_LIMIT",		TYPE_INT,		&p.blur_limit},
	{"REPERTOIRE1",		TYPE_STRING,	p.repertoire1},
	{"REPERTOIRE2",		TYPE_STRING,	p.repertoire2},
	{"COTE_RECADRAGE",	TYPE_COTE,		&p.cote_recadrage},
	{"COTE_FOND",		TYPE_COTE,		&p.cote_fond},
	{"RECADRAGE1",		TYPE_BINAIRE,	&p.isRecadrage1},
	{"FLOU",			TYPE_BINAIRE,	&p.isFlou},
	{"SAUVEGARDE1",		TYPE_BINAIRE,	&p.isSauvegarde1},
	{"ROTATION",		TYPE_BINAIRE,	&p.isRotation},
	{"RECADRAGE2",		TYPE_BINAIRE,	&p.isRecadrage2},
	{"SAUVEGARDE2",		TYPE_BINAIRE,	&p.isSauvegarde2}
};

void GetParams(int argc, _TCHAR* argv[])
{
	FILE	*pFile;
	char	pTmp[256];
	char	pTmp2[256];
	char	pConfFile[256];
	char	*pPtr;
	u32		qt_params = sizeof(g_pCmd) / sizeof(g_pCmd[0]);
	u32		i;
	float	tmp;

	// valeurs par dfaut
	p.file[0] = 0;
	strcpy(pConfFile, "config.txt");
	p.detect_width = 20;
	p.detect_diff = 25;
	p.vision_limit1 = 64;
	p.vision_limit2 = 3;
	p.vision_width1 = 16;
	p.vision_width2 = 16;
	p.vision_width3 = 128;
	p.vision_angle = 3.0;
	p.blur_radius = 3;
	p.blur_limit = 50;
	strcpy(p.repertoire1, "c:/t/");
	strcpy(p.repertoire2, "c:/trot/");
	p.cote_fond = BIT_DROITE;
	p.cote_recadrage = BIT_DROITE|BIT_BAS;
	p.isRecadrage1 = TRUE;
	p.isFlou = TRUE;
	p.isSauvegarde1 = TRUE;
	p.isRotation = TRUE;
	p.isRecadrage2 = TRUE;
	p.isSauvegarde2 = TRUE;
	p.isShowFilters = FALSE;

	// lit les paramtres de la ligne de commande
	for (i=1; i<argc; i++)
	{
		if (argv[i][0] == '/')
		{
			if (argv[i][1] == 'c')
			{
				GetRep(&argv[i][2], pConfFile);
			}
			else if (argv[i][1] == 'f')
			{
				p.isShowFilters = TRUE;
			}
			else
			{
				Error("parametre inconnu: \"%s\"\n", argv[i]);
				p.file[0] = 0;
				return;
			}
		}
		else
		{
			GetRep(argv[i], p.file);
		}
	}

	// lit le fichier de config
	pFile = fopen(pConfFile, "r");

	if (pFile == NULL)
	{
		return;
	}

	while (fgets(pTmp, 256, pFile) != NULL)
	{
		while(pTmp[strlen(pTmp)-1] == 0xA || pTmp[strlen(pTmp)-1] ==0xD)
		{
			pTmp[strlen(pTmp)-1] = 0;
		}

		pPtr = strchr(pTmp, '=');

		if (pPtr != NULL)
		{
			memcpy(pTmp2, pTmp, pPtr-pTmp);
			pTmp2[pPtr-pTmp] = 0;
			strupr(pTmp2);

			for (i=0; i<qt_params; i++)
			{
				if (strcmp(pTmp2, g_pCmd[i].pCmd) == 0)
				{
					switch (g_pCmd[i].type)
					{
						case TYPE_INT:
							sscanf(pPtr+1, "%d", g_pCmd[i].pVar);
							break;

						case TYPE_FLOAT:
							tmp = 0.0;
							sscanf(pPtr+1, "%f", &tmp);
							*((double *)g_pCmd[i].pVar) = tmp;
							break;

						case TYPE_STRING:
							GetRep(pPtr+1, (char *)g_pCmd[i].pVar);
							break;

						case TYPE_COTE:
							*((u8 *)g_pCmd[i].pVar) = 0;
							strcpy(pTmp, pPtr+1);
							strupr(pTmp);
							pPtr = pTmp;

							while (*pPtr != 0)
							{
								switch (*pPtr)
								{
									case 'H':
										*((u8 *)g_pCmd[i].pVar) |= BIT_HAUT;
										break;

									case 'B':
										*((u8 *)g_pCmd[i].pVar) |= BIT_BAS;
										break;

									case 'G':
										*((u8 *)g_pCmd[i].pVar) |= BIT_GAUCHE;
										break;

									case 'D':
										*((u8 *)g_pCmd[i].pVar) |= BIT_DROITE;
										break;
								}
								pPtr++;
							}
							break;

						case TYPE_BINAIRE:
							if (pPtr[1] == 'o' || pPtr[1] == 'O')
							{
								*((u8 *)g_pCmd[i].pVar) = TRUE;
							}
							else
							{
								*((u8 *)g_pCmd[i].pVar) = FALSE;
							}
							break;
					}
					break;
				}
			}

			if (i == qt_params)
			{
				printf("paramtre inconnu \"%s\"\n", pTmp2);
			}
		}
	}
	fclose(pFile);
}

/******************************************************************************
 *
 * Lit un pixel d'une bitmap
 *
 ******************************************************************************/
void GetPx(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	switch (p_pBm->format)
	{
		case BITMAP_GREY:
			pAddr = BITMAP_ADDR8(p_pBm, p_x, p_y);
			p_pCol->r = *pAddr;
			p_pCol->g = *pAddr;
			p_pCol->b = *pAddr;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
			p_pCol->r = *pAddr++;
			p_pCol->g = *pAddr++;
			p_pCol->b = *pAddr;
			break;
	}
}

/******************************************************************************
 *
 * Ecrit un pixel dans une bitmap
 *
 ******************************************************************************/
void SetPx(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	switch (p_pBm->format)
	{
		case BITMAP_GREY:
			*BITMAP_ADDR8(p_pBm, p_x, p_y) = p_pCol->r;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
			*pAddr++ = p_pCol->r;
			*pAddr++ = p_pCol->g;
			*pAddr   = p_pCol->b;
			break;
	}
}

/******************************************************************************
 *
 * Alloue une bitmap et sa mmoire
 *
 ******************************************************************************/
u32 AllocBm(TBitmap **p_pBitmap, u16 p_width, u16 p_height, u8 p_format)
{
	*p_pBitmap = (TBitmap *)MEM_Calloc(sizeof(TBitmap));

	if (*p_pBitmap == NULL)
	{
		return (FALSE);
	}

	if (BITMAP_Alloc(*p_pBitmap, p_width, p_height, p_format) == FALSE)
	{
		MEM_Free((void **)p_pBitmap);
		return (FALSE);
	}
	return (TRUE);
}

/******************************************************************************
 *
 * Remplace une bitmap par une autre en librant l'ancienne
 *
 ******************************************************************************/
void ReplaceBm(TBitmap **p_pOldBM, TBitmap *p_pNewBM)
{
	BITMAP_Free(*p_pOldBM);
	MEM_Free((void **)p_pOldBM);
	*p_pOldBM = p_pNewBM;
}

/******************************************************************************
 *
 * Redimensionne une Bitmap (ne marche pas avec les images en 256 couleurs!)
 *
 ******************************************************************************/
u32 Crop(TBitmap **p_pBm, s32 p_x1, s32 p_y1, s32 p_x2, s32 p_y2)
{
	TBitmap *pTmpBM;
	u16	sizex = (u16)(p_x2 - p_x1 + 1);
	u16	sizey = (u16)(p_y2 - p_y1 + 1);
	u32	y;

	if (AllocBm(&pTmpBM, sizex, sizey, (*p_pBm)->format) == FALSE)
	{
		return (FALSE);
	}

	if ((*p_pBm)->format == BITMAP_GREY)
	{
		for (y=0; y<sizey; y++)
		{
			memcpy(BITMAP_ADDR8(pTmpBM, 0, y), BITMAP_ADDR8(*p_pBm, p_x1, p_y1 + y), sizex);
		}
	}
	else
	{
		for (y=0; y<sizey; y++)
		{
			memcpy(BITMAP_ADDR24(pTmpBM, 0, y), BITMAP_ADDR24(*p_pBm, p_x1, p_y1 + y), sizex*3);
		}
	}

	ReplaceBm(p_pBm, pTmpBM);
	return (TRUE);
}

/******************************************************************************
 *
 * Trouve la dernire colonne de la page
 *
 ******************************************************************************/
s32 RecadreDroite(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (x=p_pBm->width-1; x>=0; x--)
	{
		for (y=0; y<p_pBm->height-p.detect_width; y++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx(p_pBm, x, y+i, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (y != p_pBm->height-p.detect_width)
		{
			break;
		}
	}
	return (x);
}

/******************************************************************************
 *
 * Trouve la premire colonne de la page
 *
 ******************************************************************************/
s32 RecadreGauche(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (x=0; x<p_pBm->width; x++)
	{
		for (y=0; y<p_pBm->height-p.detect_width; y++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx(p_pBm, x, y+i, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (y != p_pBm->height-p.detect_width)
		{
			break;
		}
	}
	return (x);
}

/******************************************************************************
 *
 * Trouve la dernire ligne de la page
 *
 ******************************************************************************/
s32 RecadreBas(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (y=p_pBm->height-1; y>=0; y--)
	{
		for (x=0; x<p_pBm->width-p.detect_width; x++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx(p_pBm, x+i, y, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (x != p_pBm->width-p.detect_width)
		{
			break;
		}
	}
	return (y);
}

/******************************************************************************
 *
 * Trouve la premire ligne de la page
 *
 ******************************************************************************/
s32 RecadreHaut(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width-p.detect_width; x++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx(p_pBm, x+i, y, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (x != p_pBm->width-p.detect_width)
		{
			break;
		}
	}
	return (y);
}

/******************************************************************************
 *
 * Recadre dans tous les sens
 *
 ******************************************************************************/
u32 Recadre(TBitmap **p_pBm, TColor *p_pBcol)
{
	s32	x1, y1, x2, y2;

	x1 = 0;
	y1 = 0;
	x2 = (*p_pBm)->width - 1;
	y2 = (*p_pBm)->height - 1;

	if (p.cote_recadrage & BIT_HAUT)
	{
		y1 = RecadreHaut(*p_pBm, p_pBcol);
	}

	if (p.cote_recadrage & BIT_BAS)
	{
		y2 = RecadreBas(*p_pBm, p_pBcol);
	}

	if (p.cote_recadrage & BIT_GAUCHE)
	{
		x1 = RecadreGauche(*p_pBm, p_pBcol);
	}

	if (p.cote_recadrage & BIT_DROITE)
	{
		x2 = RecadreDroite(*p_pBm, p_pBcol);
	}

	return (Crop(p_pBm, x1, y1, x2, y2));
}

/******************************************************************************
 *
 * Trouve la couleur de fond de l'image
 *
 ******************************************************************************/
void GetBGCol(TBitmap *p_pBm, TColor *p_pBcol)
{
	TColor col;
	s32	i, r, g, b, s, c;

	r = 0;
	g = 0;
	b = 0;

	if (p.cote_fond & (BIT_HAUT|BIT_BAS))
	{
		s = p_pBm->width;

		for (i=0; i<s; i++)
		{
			c = (p.cote_fond & BIT_HAUT ? 0 : p_pBm->height-1);
			GetPx(p_pBm, i, c, &col);
			r += col.r;
			g += col.g;
			b += col.b;
		}
	}
	else
	{
		s = p_pBm->height;

		for (i=0; i<s; i++)
		{
			c = (p.cote_fond & BIT_GAUCHE ? 0 : p_pBm->width-1);
			GetPx(p_pBm, c, i, &col);
			r += col.r;
			g += col.g;
			b += col.b;
		}
	}

	p_pBcol->r = (u8)(r / s);
	p_pBcol->g = (u8)(g / s);
	p_pBcol->b = (u8)(b / s);
}

/******************************************************************************
 *
 * Initialise le noyau du flou gaussien
 *
 ******************************************************************************/
s32 *Gaussian_Init(void)
{
	s32	*pBlur = (s32 *)MEM_Alloc(sizeof(int)*(2*p.blur_radius+1)*(2*p.blur_radius+1));
	s32	*pBlur2 = pBlur;
	double sigma = (double)p.blur_radius / 3.0;
	double e;
	int	x, y;

	if (pBlur != NULL)
	{
		for (y=-p.blur_radius; y<p.blur_radius+1; y++)
		{
			for (x=-p.blur_radius; x<p.blur_radius+1; x++)
			{
				e = (double)(x*x + y*y);
				e = exp(-e / (2.0*sigma*sigma));
				*pBlur2++ = (int)(e * 1024.0);
			}
		}
	}

	return (pBlur);
}

/******************************************************************************
 *
 * Applique le flou gaussien
 *
 ******************************************************************************/
u32 Gaussian_Process(TBitmap **p_pBm)
{
	s32	x, y, x2, y2;
	s32	r, g, b, k;
	s32	dr, dg, db;
	s32	*pBlur;
	s32	*pBlur2;
	TColor	colm;
	TColor	col;
	TBitmap	*pTmpBM;

	// init du flou
	pBlur = Gaussian_Init();

	if (pBlur == NULL)
	{
		return(FALSE);
	}

	if (AllocBm(&pTmpBM, (*p_pBm)->width, (*p_pBm)->height, (*p_pBm)->format) == FALSE)
	{
		MEM_Free((void **)&pBlur);
		return(FALSE);
	}

	for (y=0; y<pTmpBM->height; y++)
	{
		for (x=0; x<pTmpBM->width; x++)
		{
			GetPx(*p_pBm, x, y, &colm);
			r = 0;
			g = 0;
			b = 0;
			k = 0;
			pBlur2 = pBlur;

			for (y2=-p.blur_radius; y2<p.blur_radius+1; y2++)
			{
				for (x2=-p.blur_radius; x2<p.blur_radius+1; x2++)
				{
					if (x+x2 >= 0 && x+x2 < pTmpBM->width &&
						y+y2 >= 0 && y+y2 < pTmpBM->height)
					{
						GetPx(*p_pBm, x+x2, y+y2, &col);
						dr = (s32)colm.r - (s32)col.r;
						dg = (s32)colm.g - (s32)col.g;
						db = (s32)colm.b - (s32)col.b;

						if (dr*dr + dg*dg + db*db < p.blur_limit*p.blur_limit)
						{
							r += *pBlur2 * (s32)col.r;
							g += *pBlur2 * (s32)col.g;
							b += *pBlur2 * (s32)col.b;
							k += *pBlur2;
						}
					}
					pBlur2++;
				}
			}
			col.r = (u8)(r / k);
			col.g = (u8)(g / k);
			col.b = (u8)(b / k);
			SetPx(pTmpBM, x, y, &col);
		}
	}
	MEM_Free((void **)&pBlur);
	ReplaceBm(p_pBm, pTmpBM);
	return (TRUE);
}

char pTmpFilename[256];

/******************************************************************************
 *
 * Filtre 1: ne garde que les transitions verticales franches
 *
 ******************************************************************************/
void Filtre1(TBitmap *p_pBm)
{
	s32	x, y, c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			if (y == p_pBm->height-1)
			{
				*pPtr++ = 0;
			}
			else
			{
				c = abs((s32)pPtr[0] - (s32)pPtr[p_pBm->width]);
				*pPtr++ = (c < p.vision_limit1 ? 0 : 255);
			}
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE)
	{
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre1.tga");
		TGA_Save(pTmpFilename, p_pBm);
	}
}

/******************************************************************************
 *
 * Filtre 2: ne garde que les transition frquentes
 *
 ******************************************************************************/
void Filtre2(TBitmap *p_pBm)
{
	s32	x, y, x2, x3, c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			c = 0;

			if (x < p_pBm->width-p.vision_width1)
			{
				for (x2=0; x2<p.vision_width1-p.vision_limit2; x2++)
				{
					for (x3=0; x3<p.vision_limit2; x3++)
					{
						if (pPtr[x2+x3] == 0)
						{
							break;
						}
					}

					if (x3 == p.vision_limit2)
					{
						c++;
					}
				}
			}
			*pPtr++ = (c != 0 ? 255 : 0);
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE)
	{
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre2.tga");
		TGA_Save(pTmpFilename, p_pBm);
	}
}

/******************************************************************************
 *
 * Filtre 3: tend les lignes
 *
 ******************************************************************************/
void Filtre3(TBitmap *p_pBm)
{
	s32	x, y, x2, x3;
	u8	*pPtr;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			pPtr = BITMAP_ADDR8(p_pBm, x, y);

			if (pPtr[0] == 0 && x != 0)
			{
				if (pPtr[-1] == 255)
				{
					for (x2=0; x2<p.vision_width2; x2++)
					{
						if (x+x2 >= p_pBm->width)
						{
							x = p_pBm->width;
							break;
						}

						if (pPtr[x2] == 255 || (y != 0 && pPtr[x2-p_pBm->width] == 255) || (y != p_pBm->height-1 && pPtr[x2+p_pBm->width] == 255))
						{
							for (x3=0; x3<x2; x3++)
							{
								*pPtr++ = 1;
								x++;
							}
							pPtr++;
							x++;
							break;
						}
					}
				}
			}
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE)
	{
		TBitmap	bm;
		TColor	c;

		BITMAP_Alloc(&bm, p_pBm->width, p_pBm->height, BITMAP_GREY);

		for (y=0; y<bm.height; y++)
		{
			for (x=0; x<bm.width; x++)
			{
				GetPx(p_pBm, x, y, &c);

				if (c.r == 1)
				{
					c.r = c.g = c.b = 128;
				}
				SetPx(&bm, x, y, &c);
			}
		}
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre3.tga");
		TGA_Save(pTmpFilename, &bm);
		BITMAP_Free(&bm);
	}
}

/******************************************************************************
 *
 * Filtre 4: lignes d'1 pixel
 *
 ******************************************************************************/
void Filtre4(TBitmap *p_pBm)
{
	s32	x, y;
	u8	c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			c = 0;

			if (y < p_pBm->height-1)
			{
				if (pPtr[0] != 0 && pPtr[p_pBm->width] == 0)
				{
					c = 255;
				}
			}
			*pPtr++ = c;
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE)
	{
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre4.tga");
		TGA_Save(pTmpFilename, p_pBm);
	}
}

/******************************************************************************
 *
 * Filtre 5: moyenne des angles des lignes les plus longues
 *
 ******************************************************************************/
double Filtre5(TBitmap *p_pBm)
{
	s32	x, y, x2, y2;
	double angle = 0.0;
	double angle2, dist;
	double fac = 0.0;

	// On parcourt l'image de gauche  droite, puis de haut en bas pour trouver tous les dbuts de lignes
	for (x=0; x<p_pBm->width; x++)
	{
		for (y=0; y<p_pBm->height; y++)
		{
			if (*BITMAP_ADDR8(p_pBm, x, y) == 255)
			{
				// on a trouv le dbut d'une ligne
				x2 = x;
				y2 = y;

				// on la suit pour voir o elle s'arrte
				while (1)
				{
					// on colorie cette ligne temporairement avec la couleur 1
					*BITMAP_ADDR8(p_pBm, x2, y2) = 1;

					if (x2 >= p_pBm->width-1)
					{
						break;
					}

					if (*BITMAP_ADDR8(p_pBm, x2+1, y2) > 128)
					{
						// essaye d'aller  droite
						x2++;
					}
					else if (y2 > 0 && *BITMAP_ADDR8(p_pBm, x2+1, y2-1) > 128)
					{
						// essaye d'aller en haut  droite
						x2++;
						y2--;
					}
					else if (y2 < p_pBm->height-1 && *BITMAP_ADDR8(p_pBm, x2+1, y2+1) > 128)
					{
						// essaye d'aller en bas  droite
						x2++;
						y2++;
					}
					else
					{
						break;
					}
				}

				// est-ce que la ligne est assez longue
				if (x2 - x >= p.vision_width3)
				{
					angle2 = atan2((double)(y-y2), (double)(x2-x));
					angle2 = angle2 * 180.0 / pi;

					// est-ce que l'angle de la ligne n'est pas trop grand ?
					if (fabs(angle2) < p.vision_angle)
					{
						dist = (double)(x2-x);
						angle += dist * angle2;
						fac += dist;

						// si l'on veut sauver le rsultat, on recolorie cette ligne
						if (p.isShowFilters == TRUE)
						{
							x2 = x;
							y2 = y;

							while (1)
							{
								*BITMAP_ADDR8(p_pBm, x2, y2) = 2;

								if (x2 >= p_pBm->width-1)
								{
									break;
								}

								if (*BITMAP_ADDR8(p_pBm, x2+1, y2) == 1)
								{
									x2++;
								}
								else if (y2 > 0 && *BITMAP_ADDR8(p_pBm, x2+1, y2-1) == 1)
								{
									x2++;
									y2--;
								}
								else if (y2 < p_pBm->height-1 && *BITMAP_ADDR8(p_pBm, x2+1, y2+1) == 1)
								{
									x2++;
									y2++;
								}
								else
								{
									break;
								}
							}
						}
					}
				}
			}
		}
	}

	if (fac != 0.0)
	{
		angle = angle / fac;
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE)
	{
		TBitmap	bm;
		TColor	c;

		BITMAP_Alloc(&bm, p_pBm->width, p_pBm->height, BITMAP_GREY);

		for (y=0; y<bm.height; y++)
		{
			for (x=0; x<bm.width; x++)
			{
				GetPx(p_pBm, x, y, &c);

				if (c.r == 1)
				{
					c.r = c.g = c.b = 128;
				}
				else if (c.r == 2)
				{
					c.r = c.g = c.b = 255;
				}
				SetPx(&bm, x, y, &c);
			}
		}
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre5.tga");
		TGA_Save(pTmpFilename, &bm);
		BITMAP_Free(&bm);
	}

	return (angle);
}

/******************************************************************************
 *
 * calcule un coefficient pour l'interpolation bicubique en fonction de la distance du point
 *
 ******************************************************************************/
double Rotate_CalcCoef(double p_dist)
{
	double	dist2 = p_dist * p_dist;

	if (p_dist < 1.0)
	{
		return (1.5 * dist2*p_dist - 2.5 * dist2 + 1.0);
	}
	else if (p_dist < 2.0)
	{
		return (-0.5 * dist2*p_dist + 2.5 * dist2 - 4.0 * p_dist + 2.0);
	}
	else
	{
		return (0.0);
	}
}

/******************************************************************************
 *
 * limite une valeur entre 0 et 255
 *
 ******************************************************************************/
u8 limit(s32 p_val)
{
	if (p_val < 0)
	{
		return (0);
	}
	else if (p_val > 255)
	{
		return (255);
	}
	else
	{
		return ((u8)p_val);
	}
}

/******************************************************************************
 *
 * trouve l'image d'un point aprs la rotation, en utilisant une interpolation bicubique
 *
 ******************************************************************************/
void RotatePoint(TBitmap *p_pSrc, double p_x, double p_y, TColor *p_pBcol, TColor *p_pCol)
{
	s32	intx = (int)floor(p_x);
	s32	inty = (int)floor(p_y);
	s32	srcx, srcy;
	double	dist, coef;
	double	fr, fg, fb, fac;
	TColor	tmpc;

	fac = 0.0;
	fr = 0.0;
	fg = 0.0;
	fb = 0.0;

	for (srcx=intx-1; srcx<intx+3; srcx++)
	{
		for (srcy=inty-1; srcy<inty+3; srcy++)
		{
			dist = fabs((double)srcx + 0.5 - p_x);
			coef = Rotate_CalcCoef(dist);
			dist = fabs((double)srcy + 0.5 - p_y);
			coef *= Rotate_CalcCoef(dist);
			fac += coef;

			if (srcx >= 0 && srcx < p_pSrc->width &&
				srcy >= 0 && srcy < p_pSrc->height)
			{
				GetPx(p_pSrc, srcx, srcy, &tmpc);
				fr += coef * (double)tmpc.r;
				fg += coef * (double)tmpc.g;
				fb += coef * (double)tmpc.b;
			}
			else
			{
				fr += coef * (double)p_pBcol->r;
				fg += coef * (double)p_pBcol->g;
				fb += coef * (double)p_pBcol->b;
			}
		}
	}
	p_pCol->r = limit((s32)floor(fr / fac));
	p_pCol->g = limit((s32)floor(fg / fac));
	p_pCol->b = limit((s32)floor(fb / fac));
}

/******************************************************************************
 *
 * Rotation de l'image
 *
 ******************************************************************************/
void Rotate(TBitmap *p_pSrc, TBitmap *p_pDst, double p_angle, TColor *p_pBcol)
{
	double	center_x, center_y;
	double	src_x, src_y;
	double	savex, savey;
	double	incx, incy;
	double	angle, angle2;
	double	dist;
	s32	dst_x, dst_y;
	TColor	col;

	angle = p_angle * pi / 180.0;
	incx = cos(angle);
	incy = -sin(angle);

	center_x = (float)p_pSrc->width / 2.0 - 0.5;
	center_y = (float)p_pSrc->height / 2.0 - 0.5;
	dist = sqrt(center_x*center_x + center_y*center_y);
	angle2 = atan2(-center_y, center_x);
	savex = center_x+0.5 - dist * cos(angle2+angle);
	savey = center_y+0.5 + dist * sin(angle2+angle);

	for (dst_y=0; dst_y<p_pDst->height; dst_y++)
	{
		src_x = savex;
		src_y = savey;
		savex -= incy;
		savey += incx;

		for (dst_x=0; dst_x<p_pDst->width; dst_x++)
		{
			RotatePoint(p_pSrc, src_x, src_y, p_pBcol, &col);
			SetPx(p_pDst, dst_x, dst_y, &col);
			src_x += incx;
			src_y += incy;
		}
	}
}

/******************************************************************************
 *
 * Dtection de l'angle et rotation
 *
 ******************************************************************************/
u32 Autorotate(TBitmap **p_pBm, TColor *p_pBack_col)
{
	double	angle;
	TBitmap work_bitmap;
	TBitmap	*pTmpBM;

	// cre le buffer de travail pour la dtection de l'angle
	if (BITMAP_Alloc(&work_bitmap, (*p_pBm)->width, (*p_pBm)->height, BITMAP_GREY) == FALSE)
	{
		return (FALSE);
	}
	BITMAP_Blt(*p_pBm, &work_bitmap, 0, 0);

	// ne garde que les transitions verticales franches
	Filtre1(&work_bitmap);

	// ne garde que les transition frquentes
	Filtre2(&work_bitmap);

	// tend les lignes
	Filtre3(&work_bitmap);

	// lignes d'1 pixel
	Filtre4(&work_bitmap);

	// moyenne des angles des lignes les plus longues
	angle = Filtre5(&work_bitmap);
	printf("%f\n", angle);
	BITMAP_Free(&work_bitmap);

	// rotation
	if (AllocBm(&pTmpBM, (*p_pBm)->width, (*p_pBm)->height, (*p_pBm)->format) == FALSE)
	{
		return (FALSE);
	}

	BITMAP_Clr(pTmpBM, p_pBack_col->r, p_pBack_col->g, p_pBack_col->b);
	Rotate(*p_pBm, pTmpBM, angle, p_pBack_col);
	ReplaceBm(p_pBm, pTmpBM);
	return (TRUE);
}

/******************************************************************************
 *
 * Convertit une bitmap dans un autre format
 *
 ******************************************************************************/
u32 ConvertBm(TBitmap **p_pBm, u8 p_format)
{
	TBitmap	*pTmpBM;

	if (AllocBm(&pTmpBM, (*p_pBm)->width, (*p_pBm)->height, p_format) == FALSE)
	{
		return (FALSE);
	}
	BITMAP_Blt(*p_pBm, pTmpBM, 0, 0);
	ReplaceBm(p_pBm, pTmpBM);
	return (TRUE);
}

/******************************************************************************
 *
 * Point d'entre du programme
 *
 ******************************************************************************/
int _tmain(int argc, _TCHAR* argv[])
{
	u32		isFileLoaded;
	u32		ret;
	char	pFilename[256];
	char	*pTmp;
	TBitmap	*pBm = NULL;
	TColor	back_col;

	// init
	GetParams(argc, argv);

	if (p.file[0] == 0)
	{
		printf("\n");
		printf("retouche v2.3 par Frederic Goset\n");
		printf("  usage: retouche [flags] <fichier image>\n");
		printf("\n");
		printf("les flags possibles sont:\n");
		printf("  /c<fichier de config> : le nom du fichier config a utiliser\n");
		printf("                          exemple: retouche /c\"config2.txt\" image1.tga\n");
		printf("  /f : pour sauver les resultats des fitres dans le repertoire 2\n");
		return (0);
	}

	// alloue la bitmap de dpart
	pBm = (TBitmap *)MEM_Alloc(sizeof(TBitmap));

	// chargement
	isFileLoaded = TIFF_Load(p.file, pBm);

	if (isFileLoaded == FALSE)
	{
		isFileLoaded = TGA_Load(p.file, pBm);
	}

	if (isFileLoaded == TRUE)
	{
		ret = TRUE;

		// conversion du format
		if (pBm->format == BITMAP_RGB)
		{
			// la routine de sauvegarde TGA n'accepte pas le format RGB: on le convertit en BGR
			ret = ConvertBm(&pBm, BITMAP_BGR);
		}
		else if (pBm->format == BITMAP_256)
		{
			// le programme n'est pas conu pour les images en 256 couleurs: on les convertit en tons de gris
			ret = ConvertBm(&pBm, BITMAP_GREY);
		}

		if (ret == TRUE)
		{
			// get background color
			GetBGCol(pBm, &back_col);

			// recadrage
			if (p.isRecadrage1 == TRUE)
			{
				ret = Recadre(&pBm, &back_col);
			}

			if (ret == TRUE)
			{
				// flou gaussien
				if (p.isFlou == TRUE)
				{
					ret = Gaussian_Process(&pBm);
				}

				if (ret == TRUE)
				{
					// sauvegarde 1
					if (p.isSauvegarde1 == TRUE)
					{
						strcpy(pFilename, p.repertoire1);
						strcat(pFilename, p.file);
						pTmp = strrchr(pFilename, '.');
						strcpy(pTmp, ".tga");
						TGA_Save(pFilename, pBm);
					}

					// rotation
					if (p.isRotation == TRUE)
					{
						ret = Autorotate(&pBm, &back_col);
					}

					if (ret == TRUE)
					{
						// recadrage 2
						if (p.isRecadrage2 == TRUE)
						{
							ret = Recadre(&pBm, &back_col);
						}

						if (ret == TRUE)
						{
							// sauvegarde 2
							if (p.isSauvegarde2 == TRUE)
							{
								strcpy(pFilename, p.repertoire2);
								strcat(pFilename, p.file);
								pTmp = strrchr(pFilename, '.');
								strcpy(pTmp, ".tga");
								TGA_Save(pFilename, pBm);
							}
							printf("%s done.\n", p.file);
						}
					}
				}
			}
		}
	}

	if (pBm != NULL)
	{
		BITMAP_Free(pBm);
		MEM_Free((void **)&pBm);
	}
	return 0;
}
