
#include "stdafx.h"
#include <string.h>
#include "error.h"
#include "mem.h"
#include "bitmap.h"
#include "file.h"
#include "tiff.h"

#define	TIFF_ASSERT(_val) if (!(_val)) return (FALSE)

TTiffInfo	g_TIFF_Info;


////////////////////////////////////////////////////
//
// Routines de lecture bas niveau en prenant en compte le type d'endian
//
////////////////////////////////////////////////////
u32 TIFF_Read(u8 *p_pBuf, u32 p_size)
{
	return (FILE_Read(&g_TIFF_Info.pFile, p_pBuf, p_size, g_TIFF_Info.pFilename));
}

u32 TIFF_ReadU8(u8 *p_pVal)
{
	return (TIFF_Read(p_pVal, 1));
}

u32 TIFF_ReadU16(u16 *p_pVal)
{
	u8	pTmp[2];

	TIFF_ASSERT(TIFF_Read(pTmp, 2));

	if (g_TIFF_Info.endian == TIFF_LITTLE_ENDIAN)
	{
		*p_pVal = (pTmp[1] << 8) + pTmp[0];
	}
	else
	{
		*p_pVal = (pTmp[0] << 8) + pTmp[1];
	}
	return (TRUE);
}

u32 TIFF_ReadU32(u32 *p_pVal)
{
	u8	pTmp[4];

	TIFF_ASSERT(TIFF_Read(pTmp, 4));

	if (g_TIFF_Info.endian == TIFF_LITTLE_ENDIAN)
	{
		*p_pVal = (pTmp[3] << 24) + (pTmp[2] << 16) + (pTmp[1] << 8) + pTmp[0];
	}
	else
	{
		*p_pVal = (pTmp[0] << 24) + (pTmp[1] << 16) + (pTmp[2] << 8) + pTmp[3];
	}
	return (TRUE);
}

u32 TIFF_ReadS8(s8 *p_pVal)
{
	return(TIFF_ReadU8((u8 *)p_pVal));
}

u32 TIFF_ReadS16(s16 *p_pVal)
{
	return(TIFF_ReadU16((u16 *)p_pVal));
}

u32 TIFF_ReadS32(s32 *p_pVal)
{
	return(TIFF_ReadU32((u32 *)p_pVal));
}

u32 TIFF_ReadFloat(float *p_pVal)
{
	return(TIFF_ReadU32((u32 *)p_pVal));
}

////////////////////////////////////////////////////
//
// Routines d'criture bas niveau en prenant en compte le type d'endian
//
////////////////////////////////////////////////////
u32 TIFF_Write(u8 *p_pBuf, u32 p_size)
{
	return (FILE_Write(&g_TIFF_Info.pFile, p_pBuf, p_size, g_TIFF_Info.pFilename));
}

u32 TIFF_WriteU8(u8 *p_pVal)
{
	return (TIFF_Write(p_pVal, 1));
}

u32 TIFF_WriteU16(u16 *p_pVal)
{
	u8	pTmp[2];

	if (g_TIFF_Info.endian == TIFF_LITTLE_ENDIAN)
	{
		pTmp[0] = (u8)*p_pVal;
		pTmp[1] = (u8)(*p_pVal >> 8);
	}
	else
	{
		pTmp[0] = (u8)(*p_pVal >> 8);
		pTmp[1] = (u8)*p_pVal;
	}

	TIFF_ASSERT(TIFF_Write(pTmp, 2));
	return (TRUE);
}

u32 TIFF_WriteU32(u32 *p_pVal)
{
	u8	pTmp[4];

	if (g_TIFF_Info.endian == TIFF_LITTLE_ENDIAN)
	{
		pTmp[0] = (u8)*p_pVal;
		pTmp[1] = (u8)(*p_pVal >> 8);
		pTmp[2] = (u8)(*p_pVal >> 16);
		pTmp[3] = (u8)(*p_pVal >> 24);
	}
	else
	{
		pTmp[0] = (u8)(*p_pVal >> 24);
		pTmp[1] = (u8)(*p_pVal >> 16);
		pTmp[2] = (u8)(*p_pVal >> 8);
		pTmp[3] = (u8)*p_pVal;
	}

	TIFF_ASSERT(TIFF_Write(pTmp, 4));
	return (TRUE);
}

u32 TIFF_WriteS8(s8 *p_pVal)
{
	return(TIFF_WriteU8((u8 *)p_pVal));
}

u32 TIFF_WriteS16(s16 *p_pVal)
{
	return(TIFF_WriteU16((u16 *)p_pVal));
}

u32 TIFF_WriteS32(s32 *p_pVal)
{
	return(TIFF_WriteU32((u32 *)p_pVal));
}

u32 TIFF_WriteFloat(float *p_pVal)
{
	return(TIFF_WriteU32((u32 *)p_pVal));
}

////////////////////////////////////////////////////
//
// lit une valeur numrique dans une entre en fonction de son type
//
////////////////////////////////////////////////////
u32 TIFF_Entry_ReadInt(u16 p_type, u32 *p_pVal)
{
	u8	vu8;
	u16	vu16;

	switch (p_type)
	{
		case TIFF_ENTRYTYPE_BYTE:
			TIFF_ASSERT(TIFF_ReadU8(&vu8));
			*p_pVal = vu8;
			break;

		case TIFF_ENTRYTYPE_SHORT:
			TIFF_ASSERT(TIFF_ReadU16(&vu16));
			*p_pVal = vu16;
			break;

		case TIFF_ENTRYTYPE_LONG:
			TIFF_ASSERT(TIFF_ReadU32(p_pVal));
			break;

		default:
			return (FALSE);
	}
	return (TRUE);
}

////////////////////////////////////////////////////
//
// lit une entre
//
////////////////////////////////////////////////////
u32 TIFF_Entry_Read(void)
{
	u32	file_pos = ftell(g_TIFF_Info.pFile);	// sauve la position dans le fichier
	u16	tag, type;
	u32	count, size, offset;
	u16	data16;
	u32	data32;
	u32	i, j;

	// taille de chaque type de donnes
	static u8 const	pType_size[]=
	{
		0,	// TIFF_ENTRYTYPE_NULL
		1,	// TIFF_ENTRYTYPE_BYTE
		1,	// TIFF_ENTRYTYPE_ASCII
		2,	// TIFF_ENTRYTYPE_SHORT
		4,	// TIFF_ENTRYTYPE_LONG
		8,	// TIFF_ENTRYTYPE_RATIONAL
		1,	// TIFF_ENTRYTYPE_SBYTE
		1,	// TIFF_ENTRYTYPE_UNDEFINED
		2,	// TIFF_ENTRYTYPE_SSHORT
		4,	// TIFF_ENTRYTYPE_SLONG
		8,	// TIFF_ENTRYTYPE_SRATIONAL
		4,	// TIFF_ENTRYTYPE_FLOAT
		8	// TIFF_ENTRYTYPE_DOUBLE
	};

	// lit le tag, le type de donnes, et le nombre de donnes
	TIFF_ASSERT(TIFF_ReadU16(&tag));
	TIFF_ASSERT(TIFF_ReadU16(&type));
	TIFF_ASSERT(TIFF_ReadU32(&count));

	// saute les types de donnes inconnus
	if (type != TIFF_ENTRYTYPE_NULL && type < TIFF_ENTRYTYPE_INVALID)
	{
		// calcule la taille des donnes de cette entre
		size = pType_size[type] * count;

		// si la taille des donnes est suprieure  4, elles sont ranges  un autre endroit
		if (size > 4)
		{
			TIFF_ASSERT(TIFF_ReadU32(&offset));
			fseek(g_TIFF_Info.pFile, offset, SEEK_SET);
		}

		// interprte les diffrents tags
		switch (tag)
		{
			case TIFF_TAG_PHOTOMETRICINTERPRETATION:
				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT && count == 1);
				TIFF_ASSERT(TIFF_ReadU16(&data16));
				TIFF_ASSERT(data16 <= TIFF_PHOTO_COLORMAP);
				g_TIFF_Info.photo_interpret = (u8)data16;
				break;

			case TIFF_TAG_COMPRESSION:
				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT && count == 1);
				TIFF_ASSERT(TIFF_ReadU16(&g_TIFF_Info.compression));
				TIFF_ASSERT(g_TIFF_Info.compression == TIFF_COMPRESSION_NONE ||
						 g_TIFF_Info.compression == TIFF_COMPRESSION_HUFFMAN ||
						 g_TIFF_Info.compression == TIFF_COMPRESSION_LZW ||
						 g_TIFF_Info.compression == TIFF_COMPRESSION_PACKBITS);
				break;

			case TIFF_TAG_IMAGEWIDTH:
				TIFF_ASSERT(count == 1);
				TIFF_ASSERT(TIFF_Entry_ReadInt(type, &data32));
				g_TIFF_Info.pBm->width = (u16)data32;
				break;

			case TIFF_TAG_IMAGEHEIGHT:
				TIFF_ASSERT(count == 1);
				TIFF_ASSERT(TIFF_Entry_ReadInt(type, &data32));
				g_TIFF_Info.pBm->height = (u16)data32;
				break;

			case TIFF_TAG_STRIPOFFSETS:
				if (g_TIFF_Info.pStrips != NULL)
				{
					TIFF_ASSERT(g_TIFF_Info.qt_strips = count);
				}
				else
				{
					g_TIFF_Info.qt_strips = count;
					g_TIFF_Info.pStrips = (TTiffStrip *)MEM_Alloc(g_TIFF_Info.qt_strips * sizeof(TTiffStrip));
					TIFF_ASSERT(g_TIFF_Info.pStrips != NULL);
				}

				for (i=0; i<g_TIFF_Info.qt_strips; i++)
				{
					TIFF_ASSERT(TIFF_Entry_ReadInt(type, &g_TIFF_Info.pStrips[i].offset));
				}
				break;

			case TIFF_TAG_STRIPBYTECOUNTS:
				if (g_TIFF_Info.pStrips != NULL)
				{
					TIFF_ASSERT(g_TIFF_Info.qt_strips = count);
				}
				else
				{
					g_TIFF_Info.qt_strips = count;
					g_TIFF_Info.pStrips = (TTiffStrip *)MEM_Alloc(g_TIFF_Info.qt_strips * sizeof(TTiffStrip));
					TIFF_ASSERT(g_TIFF_Info.pStrips != NULL);
				}

				for (i=0; i<g_TIFF_Info.qt_strips; i++)
				{
					TIFF_ASSERT(TIFF_Entry_ReadInt(type, &g_TIFF_Info.pStrips[i].bytes));
				}
				break;

			case TIFF_TAG_ROWSPERSTRIP:
				TIFF_ASSERT(count == 1);
				TIFF_ASSERT(TIFF_Entry_ReadInt(type, &g_TIFF_Info.rows_per_strip));
				break;

			case TIFF_TAG_BITSPERSAMPLE:
				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT);
				TIFF_ASSERT(TIFF_ReadU16(&data16));
				TIFF_ASSERT(data16 == 1 || data16 == 4 || data16 == 8);
				g_TIFF_Info.bits_per_sample = (u8)data16;
				break;

			case TIFF_TAG_COLORMAP:
			{
				u32	qt_colors = 1 << g_TIFF_Info.bits_per_sample;

				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT && count == qt_colors*3);

				for (i=0; i<3; i++)
				{
					for (j=0; j<qt_colors; j++)
					{
						TIFF_ASSERT(TIFF_ReadU16(&data16));
						g_TIFF_Info.pPal[j][i] = data16 >> 8;
					}
				}
				break;
			}

			case TIFF_TAG_SAMPLESPERPIXEL:
				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT && count == 1);
				TIFF_ASSERT(TIFF_ReadU16(&data16));
				TIFF_ASSERT(data16 <= 3);
				g_TIFF_Info.samples_per_pixel = (u8)data16;
				break;

			case TIFF_TAG_FILLORDER:
				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT && count == 1);
				TIFF_ASSERT(TIFF_ReadU16(&data16));
				TIFF_ASSERT(data16 == TIFF_ORDER_HIGHBITSFIRST || data16 == TIFF_ORDER_LOWBITSFIRST);
				g_TIFF_Info.fill_order = (u8)data16;
				break;
	
			case TIFF_TAG_PREDICTOR:
				TIFF_ASSERT(type == TIFF_ENTRYTYPE_SHORT && count == 1);
				TIFF_ASSERT(TIFF_ReadU16(&data16));
				TIFF_ASSERT(data16 == TIFF_PREDICTOR_NONE || data16 == TIFF_PREDICTOR_DIFFERENCE);
				g_TIFF_Info.predictor = (u8)data16;
				break;
		}
	}

	// on passe  l'entre suivante grce  la position qu'on a sauv au dbut
	fseek(g_TIFF_Info.pFile, file_pos + 12, SEEK_SET);
	return (TRUE);
}

////////////////////////////////////////////////////
//
// dcompression d'un strip: PackBits
//
////////////////////////////////////////////////////
u32 TIFF_Uncomp_PackBits(u8 *p_pSrc, u8 *p_pDst, u32 p_height)
{
	u32	j, y;
	s8	cmd;
	u16	num;

	for (y=0; y<p_height; y++)
	{
		j = 0;

		while (j != g_TIFF_Info.bytes_per_row)
		{
			cmd = (s8)*p_pSrc++;

			if (cmd > 0)
			{
				num = (u16)cmd + 1;

				if (j + num > g_TIFF_Info.bytes_per_row)
				{
					return (FALSE);
				}
				memcpy(p_pDst + j, p_pSrc, num);
				j += num;
				p_pSrc += num;
			}
			else if (cmd != -128)
			{
				num = (u16)(-cmd) + 1;

				if (j + num > g_TIFF_Info.bytes_per_row)
				{
					return (FALSE);
				}
				memset(p_pDst + j, *p_pSrc++, num);
				j += num;
			}
		}
		p_pDst += g_TIFF_Info.bytes_per_row;
	}
	return (TRUE);
}

////////////////////////////////////////////////////
//
// dcompression LZW: Ajoute une chaine  la table
//
////////////////////////////////////////////////////
void TIFF_ULZW_StoreString(TTIFF_LZW *p_pInfo, u16 p_code, u8 p_char)
{
	p_pInfo->pChars[p_pInfo->table_size] = p_char;
	p_pInfo->pCodes[p_pInfo->table_size] = p_code;

	if (p_pInfo->table_size == 510)
	{
		p_pInfo->code_size = 10;
	}
	else if (p_pInfo->table_size == 1022)
	{
		p_pInfo->code_size = 11;
	}
	else if (p_pInfo->table_size == 2046)
	{
		p_pInfo->code_size = 12;
	}
	p_pInfo->table_size++;
}

////////////////////////////////////////////////////
//
// dcompression LZW: Convertit un code en chaine d'octets
//
////////////////////////////////////////////////////
u32 TIFF_ULZW_GetString(TTIFF_LZW *p_pInfo, u16 p_code, u8 *p_pString)
{
	u16	cur_code;
	u16	len, i;

	/// find string length
	len = 0;
	cur_code = p_code;

	do
	{
		len++;
		cur_code = p_pInfo->pCodes[cur_code];
	}
	while (cur_code != 0xFFFF);

	/// write string
	i = len;
	cur_code = p_code;

	do
	{
		i--;
		p_pString[i] = p_pInfo->pChars[cur_code];
		cur_code = p_pInfo->pCodes[cur_code];
	}
	while (cur_code != 0xFFFF);

	/// return string length
	return (len);
}

////////////////////////////////////////////////////
//
// dcompression LZW: initialise la table
//
////////////////////////////////////////////////////
void TIFF_ULZW_InitTable(TTIFF_LZW *p_Info)
{
	u32	i;

	for (i=0; i<256; i++)
	{
		p_Info->pCodes[i] = 0xFFFF;
		p_Info->pChars[i] = (u8)i;
	}
	p_Info->code_size = 9;
	p_Info->table_size = 258;
}

////////////////////////////////////////////////////
//
// dcompression LZW: lit le code suivant
//
////////////////////////////////////////////////////
u16 TIFF_ULZW_GetCode(u8 *p_pSrc, TTIFF_LZW *p_pInfo)
{
	u32	i;
	u16	code = 0;
	u32	ofs;

	for (i=0; i<p_pInfo->code_size; i++)
	{
		code <<= 1;
		ofs = 7 - p_pInfo->src_bit;
		code |= (p_pSrc[p_pInfo->src_off] >> ofs) & 1;
		p_pInfo->src_bit++;

		if (p_pInfo->src_bit == 8)
		{
			p_pInfo->src_bit = 0;
			p_pInfo->src_off++;
		}
	}
	return (code);
}

////////////////////////////////////////////////////
//
// dcompression d'un strip: LZW
//
////////////////////////////////////////////////////
u32 TIFF_Uncomp_LZW(u8 *p_pSrc, u8 *p_pDst, u32 p_height)
{
	u16	code, last_code;
	u8	*pTmp;
	TTIFF_LZW	info;

	info.code_size = 9;
	info.src_bit = 0;
	info.src_off = 0;
	info.dst_off = 0;

	while ((code = TIFF_ULZW_GetCode(p_pSrc, &info)) != TIFF_LZW_EOICODE)
	{
		if (code == TIFF_LZW_CLEARCODE)
		{
			TIFF_ULZW_InitTable(&info);
			code = TIFF_ULZW_GetCode(p_pSrc, &info);

			if (code == TIFF_LZW_EOICODE)
			{
				break;
			}
			info.dst_off += TIFF_ULZW_GetString(&info, code, &p_pDst[info.dst_off]);
		}
		else
		{
			pTmp = &p_pDst[info.dst_off];

			// is code in table ?
			if (code < info.table_size)
			{
				info.dst_off += TIFF_ULZW_GetString(&info, code, &p_pDst[info.dst_off]);
			}
			else
			{
				info.dst_off += TIFF_ULZW_GetString(&info, last_code, &p_pDst[info.dst_off]);
				p_pDst[info.dst_off++] = *pTmp;
			}
			TIFF_ULZW_StoreString(&info, last_code, *pTmp);
		}
		last_code = code;

if (info.dst_off >= p_height*g_TIFF_Info.bytes_per_row)	// un bug que j'ai pas trouv
{
	if (info.dst_off > p_height*g_TIFF_Info.bytes_per_row)
	{
		return (FALSE);
	}
	else
	{
		return (TRUE);
	}
}
	}
	return (TRUE);
}

////////////////////////////////////////////////////
//
// dcompression Huffman: lit le bit suivant
//
////////////////////////////////////////////////////
u8 TIFF_UHuff_ReadBit(u8 *p_pSrc, u32 *p_pOffset, u32 *p_pBit)
{
	u8		ret, mask;

	if (g_TIFF_Info.fill_order == TIFF_ORDER_HIGHBITSFIRST)
	{
		mask = 0x80 >> *p_pBit;
	}
	else
	{
		mask = 1 << *p_pBit;
	}
	ret = ((p_pSrc[*p_pOffset] & mask) == 0 ? 0 : 1) + '0';
	*p_pBit += 1;

	if (*p_pBit == 8)
	{
		*p_pBit = 0;
		*p_pOffset += 1;
	}
	return (ret);
}

////////////////////////////////////////////////////
//
// dcompression Huffman: lit le code suivant
//
////////////////////////////////////////////////////
u32 TIFF_UHuff_GetCode(u8 *p_pSrc, u32 *p_pOffset, u32 *p_pBit, u16 *p_pCode, u8 p_color)
{
	static char const	pTermCodes[2][64][TIFF_HUFF_MAXCODELEN] =
	{
		{
			/// White Terminating Codes
			"00110101",	/// 0
			"000111",	/// 1
			"0111",		/// 2
			"1000",		/// 3
			"1011",		/// 4
			"1100",		/// 5
			"1110",		/// 6
			"1111",		/// 7
			"10011",	/// 8
			"10100",	/// 9
			"00111",	/// 10
			"01000",	/// 11
			"001000",	/// 12
			"000011",	/// 13
			"110100",	/// 14
			"110101",	/// 15
			"101010",	/// 16
			"101011",	/// 17
			"0100111",	/// 18
			"0001100",	/// 19
			"0001000",	/// 20
			"0010111",	/// 21
			"0000011",	/// 22
			"0000100",	/// 23
			"0101000",	/// 24
			"0101011",	/// 25
			"0010011",	/// 26
			"0100100",	/// 27
			"0011000",	/// 28
			"00000010",	/// 29
			"00000011",	/// 30
			"00011010",	/// 31
			"00011011",	/// 32
			"00010010",	/// 33
			"00010011",	/// 34
			"00010100",	/// 35
			"00010101",	/// 36
			"00010110",	/// 37
			"00010111",	/// 38
			"00101000",	/// 39
			"00101001",	/// 40
			"00101010",	/// 41
			"00101011",	/// 42
			"00101100",	/// 43
			"00101101",	/// 44
			"00000100",	/// 45
			"00000101",	/// 46
			"00001010",	/// 47
			"00001011",	/// 48
			"01010010",	/// 49
			"01010011",	/// 50
			"01010100",	/// 51
			"01010101",	/// 52
			"00100100",	/// 53
			"00100101",	/// 54
			"01011000",	/// 55
			"01011001",	/// 56
			"01011010",	/// 57
			"01011011",	/// 58
			"01001010",	/// 59
			"01001011",	/// 60
			"00110010",	/// 61
			"00110011",	/// 62
			"00110100"	/// 63
		},
		{
			/// Black Terminating Codes
			"0000110111",	/// 0
			"010",			/// 1
			"11",			/// 2
			"10",			/// 3
			"011",			/// 4
			"0011",			/// 5
			"0010",			/// 6
			"00011",		/// 7
			"000101",		/// 8
			"000100",		/// 9
			"0000100",		/// 10
			"0000101",		/// 11
			"0000111",		/// 12
			"00000100",		/// 13
			"00000111",		/// 14
			"000011000",	/// 15
			"0000010111",	/// 16
			"0000011000",	/// 17
			"0000001000",	/// 18
			"00001100111",	/// 19
			"00001101000",	/// 20
			"00001101100",	/// 21
			"00000110111",	/// 22
			"00000101000",	/// 23
			"00000010111",	/// 24
			"00000011000",	/// 25
			"000011001010",	/// 26
			"000011001011",	/// 27
			"000011001100",	/// 28
			"000011001101",	/// 29
			"000001101000",	/// 30
			"000001101001",	/// 31
			"000001101010",	/// 32
			"000001101011",	/// 33
			"000011010010",	/// 34
			"000011010011",	/// 35
			"000011010100",	/// 36
			"000011010101",	/// 37
			"000011010110",	/// 38
			"000011010111",	/// 39
			"000001101100",	/// 40
			"000001101101",	/// 41
			"000011011010",	/// 42
			"000011011011",	/// 43
			"000001010100",	/// 44
			"000001010101",	/// 45
			"000001010110",	/// 46
			"000001010111",	/// 47
			"000001100100",	/// 48
			"000001100101",	/// 49
			"000001010010",	/// 50
			"000001010011",	/// 51
			"000000100100",	/// 52
			"000000110111",	/// 53
			"000000111000",	/// 54
			"000000100111",	/// 55
			"000000101000",	/// 56
			"000001011000",	/// 57
			"000001011001",	/// 58
			"000000101011",	/// 59
			"000000101100",	/// 60
			"000001011010",	/// 61
			"000001100110",	/// 62
			"000001100111",	/// 63
		}
	};

	static u16 const	pTermCodesLens[64] =
	{
		 0,  1,  2,  3,  4,  5,  6,  7,
		 8,  9, 10, 11, 12, 13, 14, 15,
		16, 17, 18, 19, 20, 21, 22, 23,
		24, 25, 26, 27, 28, 29, 30, 31,
		32, 33, 34, 35, 36, 37, 38, 39,
		40, 41, 42, 43, 44, 45, 46, 47,
		48, 49, 50, 51, 52, 53, 54, 55,
		56, 57, 58, 59, 60, 61, 62, 63
	};

	static char const	pMakeUpCodes[2][28][TIFF_HUFF_MAXCODELEN] =
	{
		{
			/// White Make-Up Codes
			"11011",		/// 64
			"10010",		/// 128
			"010111",		/// 192
			"0110111",		/// 256
			"00110110",		/// 320
			"00110111",		/// 384
			"01100100",		/// 448
			"01100101",		/// 512
			"01101000",		/// 576
			"01100111",		/// 640
			"011001100",	/// 704
			"011001101",	/// 768
			"011010010",	/// 832
			"011010011",	/// 896
			"011010100",	/// 960
			"011010101",	/// 1024
			"011010110",	/// 1088
			"011010111",	/// 1152
			"011011000",	/// 1216
			"011011001",	/// 1280
			"011011010",	/// 1344
			"011011011",	/// 1408
			"010011000",	/// 1472
			"010011001",	/// 1536
			"010011010",	/// 1600
			"011000",		/// 1664
			"010011011",	/// 1728

			"000000000001"	/// EOL
		},
		{
			/// Black Make-Up Codes
			"0000001111",		/// 64
			"000011001000",		/// 128
			"000011001001",		/// 192
			"000001011011",		/// 256
			"000000110011",		/// 320
			"000000110100",		/// 384
			"000000110101",		/// 448
			"0000001101100",	/// 512
			"0000001101101",	/// 576
			"0000001001010",	/// 640
			"0000001001011",	/// 704
			"0000001001100",	/// 768
			"0000001001101",	/// 832
			"0000001110010",	/// 896
			"0000001110011",	/// 960
			"0000001110100",	/// 1024
			"0000001110101",	/// 1088
			"0000001110110",	/// 1152
			"0000001110111",	/// 1216
			"0000001010010",	/// 1280
			"0000001010011",	/// 1344
			"0000001010100",	/// 1408
			"0000001010101",	/// 1472
			"0000001011010",	/// 1536
			"0000001011011",	/// 1600
			"0000001100100",	/// 1664
			"0000001100101",	/// 1728

			"00000000000"		/// EOL
		}
	};

	static u16 const	pMakeUpCodesLens[28] =
	{
		  64,  128,  192,  256,  320,  384,  448,  512,
		 576,  640,  704,  768,  832,  896,  960, 1024,
		1088, 1152, 1216, 1280, 1344, 1408, 1472, 1536,
		1600, 1664, 1728,    0
	};

	static char const	pAdditMakeUpCodes[13][TIFF_HUFF_MAXCODELEN] =
	{
		"00000001000",	/// 1792
		"00000001100",	/// 1856
		"00000001101",	/// 1920
		"000000010010",	/// 1984
		"000000010011",	/// 2048
		"000000010100",	/// 2112
		"000000010101",	/// 2176
		"000000010110",	/// 2240
		"000000010111",	/// 2304
		"000000011100",	/// 2368
		"000000011101",	/// 2432
		"000000011110",	/// 2496
		"000000011111"	/// 2560
	};

	static u16 const	pAdditMakeUpCodesLens[13] =
	{
		1792, 1856, 1920, 1984, 2048, 2112, 2176, 2240,
		2304, 2368, 2432, 2496, 2560
	};

	char	pCode[TIFF_HUFF_MAXCODELEN+1];
	u8		code_len = 0;
	u32	i;

	while(code_len < TIFF_HUFF_MAXCODELEN)
	{
		pCode[code_len++] = TIFF_UHuff_ReadBit(p_pSrc, p_pOffset, p_pBit);
		pCode[code_len] = 0;

		/// is it a term code ?
		for (i=0; i<64; i++)
		{
			if (strcmp(pCode, pTermCodes[p_color][i]) == 0)
			{
				*p_pCode = pTermCodesLens[i];
				return (TRUE);
			}
		}

		/// is it a makeup code ?
		for (i=0; i<28; i++)
		{
			if (strcmp(pCode, pMakeUpCodes[p_color][i]) == 0)
			{
				*p_pCode = pMakeUpCodesLens[i];
				return (TRUE);
			}
		}

		/// is it an additionnal makeup code ?
		for (i=0; i<13; i++)
		{
			if (strcmp(pCode, pAdditMakeUpCodes[i]) == 0)
			{
				*p_pCode = pAdditMakeUpCodesLens[i];
				return (TRUE);
			}
		}
	}
	return (FALSE);
}

////////////////////////////////////////////////////
//
// dcompression d'un strip: Huffman
//
////////////////////////////////////////////////////
u32 TIFF_Uncomp_Huffman(u8 *p_pSrc, u8 *p_pDst, u32 p_height)
{
	u32	y;
	u32	src_offset = 0, src_bit = 0, dst_offset, code_len, row_len;
	u8	color, color2;
	u16	code_val;

	for (y=0; y<p_height; y++)
	{
		color = TIFF_HUFF_WHITE;
		dst_offset = 0;
		code_len = 0;
		row_len = 0;

		while (row_len < g_TIFF_Info.pBm->width)
		{
			TIFF_ASSERT(TIFF_UHuff_GetCode(p_pSrc, &src_offset, &src_bit, &code_val, color));
			code_len += code_val;

			// Term Code ?
			if (code_val < 64)
			{
				row_len += code_len;

				if (row_len > g_TIFF_Info.pBm->width)
				{
					return (FALSE);
				}

				while(code_len)
				{
					color2 = color ^ (g_TIFF_Info.photo_interpret == TIFF_PHOTO_INVERTGRAY ? 1 : 0);
					p_pDst[dst_offset >> 3] |= color << (7 - (dst_offset & 7));
					dst_offset++;
					code_len--;
				}
				color ^= 1;
			}
		}

		// skips the last bits of the input line
		if (src_bit != 0)
		{
			src_bit = 0;
			src_offset++;
		}
	}
	return (TRUE);
}

////////////////////////////////////////////////////
//
// dcompression d'un strip: prdicteur diffrence
//
////////////////////////////////////////////////////
u8 TIFF_GetPix(u8 *p_pBuf, u32 p_x, u32 p_y)
{
	u8 *pAddr = &p_pBuf[p_y * g_TIFF_Info.bytes_per_row];

	switch (g_TIFF_Info.bits_per_sample)
	{
		case 8:
			return (pAddr[p_x]);

		case 4:
			if (p_x & 0)
			{
				return (pAddr[p_x/2] >> 4);
			}
			else
			{
				return (pAddr[p_x/2] & 0xF);
			}

		case 1:
			return ((pAddr[p_x/8] >> (7 - (p_x & 7))) & 1);
	}
	return (0);
}

void TIFF_SetPix(u8 *p_pBuf, u32 p_x, u32 p_y, u8 p_val)
{
	u8	*pAddr = &p_pBuf[p_y * g_TIFF_Info.bytes_per_row];
	u32	ofs;

	switch (g_TIFF_Info.bits_per_sample)
	{
		case 8:
			pAddr[p_x] = p_val;
			break;

		case 4:
			p_val &= 0xf;

			if (p_x & 0)
			{
				pAddr[p_x] &= 0xF;
				pAddr[p_x] |= p_val << 4;
			}
			else
			{
				pAddr[p_x] &= 0xF0;
				pAddr[p_x] |= p_val & 0xF;
			}
			break;

		case 1:
			p_val &= 1;
			ofs = 7 - (p_x & 7);
			pAddr[p_x/8] &= ~(1 << ofs);
			pAddr[p_x/8] |= p_val << ofs;
			break;
	}
}

void TIFF_UnComp_Difference(u8 *p_pBuf, u32 p_height)
{
	u32	x, y, i, p1, p2;

	for (y=0; y<p_height; y++)
	{
		for (x=0; x<(u32)(g_TIFF_Info.pBm->width-1); x++)
		{
			for (i=0; i<g_TIFF_Info.samples_per_pixel; i++)
			{
				p1 = TIFF_GetPix(p_pBuf,     x*g_TIFF_Info.samples_per_pixel+i, y);
				p2 = TIFF_GetPix(p_pBuf, (x+1)*g_TIFF_Info.samples_per_pixel+i, y);
				TIFF_SetPix(p_pBuf, (x+1)*g_TIFF_Info.samples_per_pixel+i, y, (u8)(p1+p2));
			}
		}
	}
}

////////////////////////////////////////////////////
//
// convertit un strip
//
////////////////////////////////////////////////////
u32 TIFF_Strip_Convert(u32 p_num, u32 p_height)
{
	u8	*pSrc = NULL;
	u8	*pDst = NULL;
	u32	ret, uncomp_size, x, x2, y, y2, i;
	u8	coul;

	// alloue la mmoire pour les donnes compresses
	pSrc = (u8 *)MEM_Alloc(g_TIFF_Info.pStrips[p_num].bytes);
	TIFF_ASSERT(pSrc != NULL);

	// lit les donnes compresses du fichier
	fseek(g_TIFF_Info.pFile, g_TIFF_Info.pStrips[p_num].offset, SEEK_SET);
	ret = FILE_Read(&g_TIFF_Info.pFile, pSrc, g_TIFF_Info.pStrips[p_num].bytes, g_TIFF_Info.pFilename);

	// alloue la mmoire pour les donnes dcompresses
	uncomp_size = g_TIFF_Info.bytes_per_row * p_height;
	pDst = (u8 *)MEM_Calloc(uncomp_size);

	// dcompresse les donnes
	if (ret == TRUE)
	{
		switch (g_TIFF_Info.compression)
		{
			case TIFF_COMPRESSION_NONE:
				memcpy(pDst, pSrc, g_TIFF_Info.bytes_per_row * p_height);
				break;

			case TIFF_COMPRESSION_HUFFMAN:
				ret = TIFF_Uncomp_Huffman(pSrc, pDst, p_height);
				break;

			case TIFF_COMPRESSION_LZW:
				ret = TIFF_Uncomp_LZW(pSrc, pDst, p_height);
				break;

			case TIFF_COMPRESSION_PACKBITS:
				ret = TIFF_Uncomp_PackBits(pSrc, pDst, p_height);
				break;

			default:
				ret = FALSE;
				break;
		}

		if (ret == TRUE && g_TIFF_Info.predictor == TIFF_PREDICTOR_DIFFERENCE)
		{
			TIFF_UnComp_Difference(pDst, p_height);
		}
	}
	MEM_Free((void **)&pSrc);

	// ecrit les donnes dans la bitmap
	if (ret == TRUE)
	{
		for (y=0; y<p_height; y++)
		{
			for (x=0; x<g_TIFF_Info.pBm->width; x++)
			{
				for (i=0; i<g_TIFF_Info.samples_per_pixel; i++)
				{
					x2 = x*g_TIFF_Info.samples_per_pixel+i;
					coul = TIFF_GetPix(pDst, x2, y);

					if (g_TIFF_Info.bits_per_sample == 1)
					{
						coul = (coul == 0 ? 0 : 255);
					}
					else if (g_TIFF_Info.bits_per_sample == 4)
					{
						coul = (coul << 4) + coul;
					}

					if ((g_TIFF_Info.compression == TIFF_COMPRESSION_NONE) && (g_TIFF_Info.photo_interpret == TIFF_PHOTO_INVERTGRAY))
					{
						coul = 255 - coul;
					}

					y2 = g_TIFF_Info.rows_per_strip * p_num + y;

					if (g_TIFF_Info.samples_per_pixel == 1)
					{
						*BITMAP_ADDR8(g_TIFF_Info.pBm, x2, y2) = coul;
					}
					else
					{
						BITMAP_ADDR24(g_TIFF_Info.pBm, x, y2)[i] = coul;
					}
				}
			}
		}
	}

	MEM_Free((void **)&pDst);
	return (ret);
}

////////////////////////////////////////////////////
//
// convertit tous les strips d'un directory
//
////////////////////////////////////////////////////
u32 TIFF_Dir_Convert(void)
{
	u32	i, height;

	// calcule le nombre d'octets par ligne (utilis pour la dcompression)
	g_TIFF_Info.bytes_per_row = g_TIFF_Info.pBm->width * g_TIFF_Info.samples_per_pixel * g_TIFF_Info.bits_per_sample;
	g_TIFF_Info.bytes_per_row = (g_TIFF_Info.bytes_per_row + 7) / 8;

	for (i=0; i<g_TIFF_Info.qt_strips; i++)
	{
		if (i == g_TIFF_Info.qt_strips-1)
		{
			height = g_TIFF_Info.pBm->height - (g_TIFF_Info.qt_strips - 1) * g_TIFF_Info.rows_per_strip;
		}
		else
		{
			height = g_TIFF_Info.rows_per_strip;
		}
		TIFF_ASSERT(TIFF_Strip_Convert(i, height));
	}

	return (TRUE);
}

////////////////////////////////////////////////////
//
// lit un directory
//
////////////////////////////////////////////////////
u32 TIFF_Dir_Read(u32 p_offset)
{
	u16	qt_entries;
	u32	i;
	u8	format;

	// Valeurs par dfaut (si les tags sont omis)
	g_TIFF_Info.bits_per_sample = 1;
	g_TIFF_Info.compression = TIFF_COMPRESSION_NONE;
	g_TIFF_Info.fill_order = TIFF_ORDER_HIGHBITSFIRST;
	g_TIFF_Info.rows_per_strip = 0xFFFFFFFF;
	g_TIFF_Info.samples_per_pixel = 1;

	g_TIFF_Info.pStrips = NULL;
	g_TIFF_Info.qt_strips = 0;
	g_TIFF_Info.predictor = TIFF_PREDICTOR_NONE;

	// on se positionne au dbut du directory dans le fichier
	fseek(g_TIFF_Info.pFile, p_offset, SEEK_SET);

	// trouve le nombre d'entres
	TIFF_ASSERT(TIFF_ReadU16(&qt_entries));

	// lit toutes les entres
	for (i=0; i<qt_entries; i++)
	{
		TIFF_ASSERT(TIFF_Entry_Read());
	}
	TIFF_ASSERT(g_TIFF_Info.predictor != TIFF_PREDICTOR_DIFFERENCE || g_TIFF_Info.bits_per_sample == 8);

	// trouve le format de la bitmap et alloue sa mmoire
	switch (g_TIFF_Info.photo_interpret)
	{
		case TIFF_PHOTO_INVERTGRAY:
		case TIFF_PHOTO_GRAY:
			format = BITMAP_GREY;
			break;

		case TIFF_PHOTO_RGB:
			format = BITMAP_RGB;
			break;

		case TIFF_PHOTO_COLORMAP:
			format = BITMAP_256;
			break;
	}
	TIFF_ASSERT(BITMAP_Alloc(g_TIFF_Info.pBm, g_TIFF_Info.pBm->width, g_TIFF_Info.pBm->height, format) == TRUE);

	// dcode l'image
	TIFF_ASSERT(TIFF_Dir_Convert());

	// recopie la palette
	if (g_TIFF_Info.pBm->format == BITMAP_256)
	{
		memcpy(g_TIFF_Info.pBm->pPalette, g_TIFF_Info.pPal, 256*3);
	}

/// Directory suivant
	return(TRUE);
}

////////////////////////////////////////////////////
//
// lit un fichier TIFF
//
////////////////////////////////////////////////////
#define TIFF_ERROR	BITMAP_Free(p_pBitmap); \
					FILE_Close(&g_TIFF_Info.pFile); \
					return (FALSE);

u32 TIFF_Load(char *p_pFilename, TBitmap *p_pBitmap)
{
	u8		pBuf[2];
	u16		magic;
	u32		offset;

	g_TIFF_Info.pBm = p_pBitmap;
	BITMAP_Init(g_TIFF_Info.pBm, 0, 0, 0);
	strcpy(g_TIFF_Info.pFilename, p_pFilename);

	// ouvre le fichier
	g_TIFF_Info.pFile = FILE_Open(g_TIFF_Info.pFilename, "rb");

	if (g_TIFF_Info.pFile == NULL)
	{
		TIFF_ERROR;
	}

	// lecture de l'endian
	if (FILE_Read(&g_TIFF_Info.pFile, pBuf, 2, g_TIFF_Info.pFilename) == FALSE)
	{
		TIFF_ERROR;
	}

	if (pBuf[0] == 'M' && pBuf[1] == 'M')	// Motorola
	{
		g_TIFF_Info.endian = TIFF_BIG_ENDIAN;
	}
	else if (pBuf[0] == 'I' && pBuf[1] == 'I')	// Intel
	{
		g_TIFF_Info.endian = TIFF_LITTLE_ENDIAN;
	}
	else
	{
		TIFF_ERROR;
	}

	// lit le magic number
	if (TIFF_ReadU16(&magic) == FALSE)
	{
		TIFF_ERROR;
	}

	if (magic != 42)
	{
		TIFF_ERROR;
	}

	// lit l'offset du premier directory
	if (TIFF_ReadU32(&offset) == FALSE)
	{
		TIFF_ERROR;
	}

	// lit le directory
	if (TIFF_Dir_Read(offset) == FALSE)
	{
		MEM_Free((void **)&g_TIFF_Info.pStrips);
		TIFF_ERROR;
	}

	MEM_Free((void **)&g_TIFF_Info.pStrips);
	FILE_Close(&g_TIFF_Info.pFile);
	return (TRUE);
}



#define TIFF_WERROR(_val)	if (!(_val)) \
							{ \
								FILE_Close(&g_TIFF_Info.pFile); \
								return (FALSE); \
							}

////////////////////////////////////////////////////
//
// crit une entre simple de l'IFD d'un fichier TIFF
//
////////////////////////////////////////////////////
u32 TIFF_WriteEntry(u16 p_tag, u16 p_type, u32 p_val)
{
	u8	v8;
	u16	v16;
	u32	v32 = 1;

	TIFF_WERROR(TIFF_WriteU16(&p_tag));
	TIFF_WERROR(TIFF_WriteU16(&p_type));
	TIFF_WERROR(TIFF_WriteU32(&v32));
	v32 = 0;

	switch (p_type)
	{
		case TIFF_ENTRYTYPE_BYTE:
		case TIFF_ENTRYTYPE_SBYTE:
			v8 = (u8)p_val;
			TIFF_WERROR(TIFF_WriteU8(&v8));
			TIFF_WERROR(TIFF_Write((u8 *)&v32, 3));
			break;

		case TIFF_ENTRYTYPE_SHORT:
		case TIFF_ENTRYTYPE_SSHORT:
			v16 = (u16)p_val;
			TIFF_WERROR(TIFF_WriteU16(&v16));
			TIFF_WERROR(TIFF_Write((u8 *)&v32, 2));
			break;

		case TIFF_ENTRYTYPE_LONG:
		case TIFF_ENTRYTYPE_SLONG:
			TIFF_WERROR(TIFF_WriteU32(&p_val));
			TIFF_WERROR(TIFF_Write((u8 *)&v32, 1));
			break;
	}
}

////////////////////////////////////////////////////
//
// sauve un fichier TIFF en noir et blanc
//
////////////////////////////////////////////////////
u32 TIFF_Save(char *p_pFilename, TBitmap *p_pBitmap)
{
	u16	v16;
	u32	v32, i, s, y, x, my, offs, bit;
	u32	bytesperrow;
	u32	rowsperstrip;
	u32	stripbytecount;
	u32	stripsperimage;
	u32	off_nb_entries;
	u32	off_stripoffsets;
	u32	off_stripbytecounts;
	u32	off_xres;
	u32	off_yres;
	u8	*pMem;
	u8	*pImg;

	TIFF_ASSERT(p_pBitmap->format == BITMAP_BW || p_pBitmap->format == BITMAP_GREY);
	TIFF_ASSERT(p_pBitmap->width != 0 && p_pBitmap->height != 0);
	TIFF_ASSERT(p_pFilename[0] != 0);

	bytesperrow = (p_pBitmap->width + 7) / 8;
	rowsperstrip = (8 * 1024) / bytesperrow;

	if (rowsperstrip == 0)
	{
		rowsperstrip = 1;
	}
	stripbytecount = bytesperrow * rowsperstrip;
	stripsperimage = (p_pBitmap->height + rowsperstrip-1) / rowsperstrip;

	strcpy(g_TIFF_Info.pFilename, p_pFilename);
	g_TIFF_Info.endian = TIFF_LITTLE_ENDIAN;

	// ouvre le fichier
	g_TIFF_Info.pFile = FILE_Open(g_TIFF_Info.pFilename, "wb");

	if (g_TIFF_Info.pFile == NULL)
	{
		return (FALSE);
	}

	v16 = ('I' << 8) + 'I';		// endian
	TIFF_WERROR(TIFF_WriteU16(&v16));
	v16 = 42;					// magic number
	TIFF_WERROR(TIFF_WriteU16(&v16));
	v32 = 8;					// offset de l'IFD
	TIFF_WERROR(TIFF_WriteU32(&v32));

	// nombre d'entres dans l'IFD
	off_nb_entries = ftell(g_TIFF_Info.pFile);
	v16 = 0;
	TIFF_WERROR(TIFF_WriteU16(&v16));

	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_IMAGEWIDTH, TIFF_ENTRYTYPE_SHORT, p_pBitmap->width));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_IMAGEHEIGHT, TIFF_ENTRYTYPE_SHORT, p_pBitmap->height));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_BITSPERSAMPLE, TIFF_ENTRYTYPE_SHORT, 1));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_COMPRESSION, TIFF_ENTRYTYPE_SHORT, TIFF_COMPRESSION_NONE));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_PHOTOMETRICINTERPRETATION, TIFF_ENTRYTYPE_SHORT, TIFF_PHOTO_GRAY));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_FILLORDER, TIFF_ENTRYTYPE_SHORT, TIFF_ORDER_HIGHBITSFIRST));

	v16 = TIFF_TAG_STRIPOFFSETS;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v16 = TIFF_ENTRYTYPE_LONG;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v32 = stripsperimage;
	TIFF_ASSERT(TIFF_WriteU32(&v32));
	off_stripoffsets = ftell(g_TIFF_Info.pFile);
	v32 = 0;
	TIFF_ASSERT(TIFF_WriteU32(&v32));

	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_SAMPLESPERPIXEL, TIFF_ENTRYTYPE_SHORT, 1));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_ROWSPERSTRIP, TIFF_ENTRYTYPE_SHORT, rowsperstrip));

	v16 = TIFF_TAG_STRIPBYTECOUNTS;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v16 = TIFF_ENTRYTYPE_LONG;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v32 = stripsperimage;
	TIFF_ASSERT(TIFF_WriteU32(&v32));
	off_stripbytecounts = ftell(g_TIFF_Info.pFile);
	v32 = 0;
	TIFF_ASSERT(TIFF_WriteU32(&v32));

	v16 = TIFF_TAG_XRESOLUTION;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v16 = TIFF_ENTRYTYPE_RATIONAL;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v32 = 1;
	TIFF_ASSERT(TIFF_WriteU32(&v32));
	off_xres = ftell(g_TIFF_Info.pFile);
	v32 = 0;
	TIFF_ASSERT(TIFF_WriteU32(&v32));

	v16 = TIFF_TAG_YRESOLUTION;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v16 = TIFF_ENTRYTYPE_RATIONAL;
	TIFF_ASSERT(TIFF_WriteU16(&v16));
	v32 = 1;
	TIFF_ASSERT(TIFF_WriteU32(&v32));
	off_yres = ftell(g_TIFF_Info.pFile);
	v32 = 0;
	TIFF_ASSERT(TIFF_WriteU32(&v32));

	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_RESOLUTIONUNIT, TIFF_ENTRYTYPE_SHORT, 2));
	TIFF_ASSERT(TIFF_WriteEntry(TIFF_TAG_PREDICTOR, TIFF_ENTRYTYPE_SHORT, TIFF_PREDICTOR_NONE));

	// remonte pour crire le nombre d'entres de cet IFD
	v16 = (u16)((ftell(g_TIFF_Info.pFile) - off_nb_entries - 2) / 12);
	fseek(g_TIFF_Info.pFile, off_nb_entries, SEEK_SET);
	TIFF_WERROR(TIFF_WriteU16(&v16));
	fseek(g_TIFF_Info.pFile, 0, SEEK_END);

	// offset de l'IFD suivant (0 pour dire qu'il n'y en a pas)
	v32 = 0;
	TIFF_WERROR(TIFF_WriteU32(&v32));

	// remonte pour crire l'offset de la rsolution horizontale
	v32 = ftell(g_TIFF_Info.pFile);
	fseek(g_TIFF_Info.pFile, off_xres, SEEK_SET);
	TIFF_WERROR(TIFF_WriteU32(&v32));
	fseek(g_TIFF_Info.pFile, 0, SEEK_END);

	// rsolution horizontale
	v32 = 300;
	TIFF_WERROR(TIFF_WriteU32(&v32));
	v32 = 1;
	TIFF_WERROR(TIFF_WriteU32(&v32));

	// remonte pour crire l'offset de la rsolution verticale
	v32 = ftell(g_TIFF_Info.pFile);
	fseek(g_TIFF_Info.pFile, off_yres, SEEK_SET);
	TIFF_WERROR(TIFF_WriteU32(&v32));
	fseek(g_TIFF_Info.pFile, 0, SEEK_END);

	// rsolution verticale
	v32 = 300;
	TIFF_WERROR(TIFF_WriteU32(&v32));
	v32 = 1;
	TIFF_WERROR(TIFF_WriteU32(&v32));

	// laisse de la place pour les offsets
	if (stripsperimage > 1)
	{
		v32 = ftell(g_TIFF_Info.pFile);
		fseek(g_TIFF_Info.pFile, off_stripoffsets, SEEK_SET);
		TIFF_WERROR(TIFF_WriteU32(&v32));
		fseek(g_TIFF_Info.pFile, 0, SEEK_END);
		off_stripoffsets = v32;
		v32 = 0;

		for (i=0; i<stripsperimage; i++)
		{
			TIFF_WERROR(TIFF_WriteU32(&v32));
		}
	}

	// crit les stripbytecounts
	if (stripsperimage == 1)
	{
		fseek(g_TIFF_Info.pFile, off_stripbytecounts, SEEK_SET);
		v32 = bytesperrow * p_pBitmap->height;
		TIFF_WERROR(TIFF_WriteU32(&v32));
		fseek(g_TIFF_Info.pFile, 0, SEEK_END);
	}
	else
	{
		v32 = ftell(g_TIFF_Info.pFile);
		fseek(g_TIFF_Info.pFile, off_stripbytecounts, SEEK_SET);
		TIFF_WERROR(TIFF_WriteU32(&v32));
		fseek(g_TIFF_Info.pFile, 0, SEEK_END);

		for (i=0; i<stripsperimage-1; i++)
		{
			TIFF_WERROR(TIFF_WriteU32(&stripbytecount));
		}
		v32 = p_pBitmap->height % rowsperstrip;

		if (v32 == 0)
		{
			v32 = rowsperstrip;
		}
		v32 *= bytesperrow;
		TIFF_WERROR(TIFF_WriteU32(&v32));
	}

	// alloue de la mmoire pour une ligne
	pMem = (u8 *)MEM_Alloc(bytesperrow);
	TIFF_WERROR(pMem != NULL);

	// crit les donnes
	pImg = p_pBitmap->pMem;

	for (s=0; s<stripsperimage; s++)
	{
		v32 = ftell(g_TIFF_Info.pFile);
		fseek(g_TIFF_Info.pFile, off_stripoffsets, SEEK_SET);
		TIFF_WERROR(TIFF_WriteU32(&v32));
		fseek(g_TIFF_Info.pFile, 0, SEEK_END);
		off_stripoffsets += 4;

		if (s != stripsperimage-1)
		{
			my = rowsperstrip;
		}
		else
		{
			my = p_pBitmap->height - rowsperstrip * (stripsperimage-1);
		}

		for (y=0; y<my; y++)
		{
			memset(pMem, 0, bytesperrow);
			offs = 0;
			bit = 7;

			for (x=0; x<p_pBitmap->width; x++)
			{
				if (*pImg++ > 127)
				{
					pMem[offs] |= 1 << bit;
				}

				if (bit == 0)
				{
					bit = 7;
					offs++;
				}
				else
				{
					bit--;
				}
			}
			TIFF_WERROR(TIFF_Write(pMem, bytesperrow));
		}
	}

	// ferme le fichier
	FILE_Close(&g_TIFF_Info.pFile);
	return (TRUE);
}
