// retouche.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <windows.h> /* For the CreateThread prototype */
#include "mem.h"
#include "bitmap.h"
#include "error.h"
#include "tga.h"
#include "tiff.h"

typedef struct
{
	u8	r, g, b;
} TColor;

#define	BIT_HAUT	1
#define	BIT_BAS		2
#define	BIT_GAUCHE	4
#define	BIT_DROITE	8

#define MAX_PATH	260

#define NUM_THREAD	8

struct SParams
{
	int		detect_width;
	int		detect_diff;
	int		blur_radius;
	int		blur_limit;
	int		vision_limit1;
	int		vision_limit2;
	int		vision_width1;
	int		vision_width2;
	int		vision_width3;
	u8		cote_recadrage;
	u8		cote_fond;
	u8		isRecadrage1;
	u8		isFlou;
	u8		isSauvegarde1;
	u8		isRotation;
	u8		isRecadrage2;
	u8		isSauvegarde2;
	u8		isShowFilters;
	u8		isSaveBlacknWhite;
	u8		isAutoColor;
	double	vision_angle;
	double	offset_angle;
	char	repertoire1[MAX_PATH];
	char	repertoire2[MAX_PATH];
	char	file[MAX_PATH];
	char	srcdir[MAX_PATH];
	TColor	coul_remplissage;
	u32		d_bits;
};

struct SParams	p;
double	pi = 3.14159265359;
char pTmpFilename[MAX_PATH];

void dsave(u32 p_bit, TBitmap *p_pBm, u32 p_convert);
u32 chk_bits(u32 p_bit);

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Spare le path d'un nom de fichier
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void SplitPath(char *p_pFile, char *p_pPath)
{
	char	*pChr;
	char	*pChr2;

	strcpy(p_pPath, p_pFile);
	pChr = strrchr(p_pPath, '\\');
	pChr2 = strrchr(p_pPath, ':');

	if (pChr2 > pChr)
	{
		pChr = pChr2;
	}

	pChr2 = strrchr(p_pPath, '/');

	if (pChr2 > pChr)
	{
		pChr = pChr2;
	}

	if (pChr == NULL)
	{
		*p_pPath = 0;
	}
	else
	{
		strcpy(p_pFile, pChr+1);
		*(pChr+1) = 0;
	}
}

/******************************************************************************
 *
 * Recopie une chaine de caractres en enlevant les espaces  la fin et les guillemets
 *
 ******************************************************************************/
void GetRep(char *p_pSrc, char *p_pDst)
{
	if (*p_pSrc == '\"')
	{
		p_pSrc++;
	}

	strcpy(p_pDst, p_pSrc);

	while (p_pDst[strlen(p_pDst)-1] == ' ')
	{
		p_pDst[strlen(p_pDst)-1] = 0;
	}

	if (p_pDst[strlen(p_pDst)-1] == '\"')
	{
		p_pDst[strlen(p_pDst)-1] = 0;
	}
}

/******************************************************************************
 *
 * Lit les paramtres dans le fichier de config
 *
 ******************************************************************************/
enum
{
	TYPE_INT=0,
	TYPE_FLOAT,
	TYPE_STRING,
	TYPE_COTE,
	TYPE_BINAIRE,
	TYPE_COULEUR
};

typedef struct SPrm
{
	char	pCmd[40];
	int		type;
	void	*pVar;
} TPrm;

static TPrm g_pCmd[] =
{
	{"DETECT_WIDTH",		TYPE_INT,		&p.detect_width},
	{"DETECT_DIFF",			TYPE_INT,		&p.detect_diff},
	{"VISION_LIMIT1",		TYPE_INT,		&p.vision_limit1},
	{"VISION_LIMIT2",		TYPE_INT,		&p.vision_limit2},
	{"VISION_WIDTH1",		TYPE_INT,		&p.vision_width1},
	{"VISION_WIDTH2",		TYPE_INT,		&p.vision_width2},
	{"VISION_WIDTH3",		TYPE_INT,		&p.vision_width3},
	{"VISION_ANGLE",		TYPE_FLOAT,		&p.vision_angle},
	{"BLUR_RADIUS",			TYPE_INT,		&p.blur_radius},
	{"BLUR_LIMIT",			TYPE_INT,		&p.blur_limit},
	{"REPERTOIRE1",			TYPE_STRING,	p.repertoire1},
	{"REPERTOIRE2",			TYPE_STRING,	p.repertoire2},
	{"COTE_RECADRAGE",		TYPE_COTE,		&p.cote_recadrage},
	{"COTE_FOND",			TYPE_COTE,		&p.cote_fond},
	{"OFFSET_ANGLE",		TYPE_FLOAT,		&p.offset_angle},
	{"RECADRAGE1",			TYPE_BINAIRE,	&p.isRecadrage1},
	{"FLOU",				TYPE_BINAIRE,	&p.isFlou},
	{"SAUVEGARDE1",			TYPE_BINAIRE,	&p.isSauvegarde1},
	{"ROTATION",			TYPE_BINAIRE,	&p.isRotation},
	{"RECADRAGE2",			TYPE_BINAIRE,	&p.isRecadrage2},
	{"SAUVEGARDE2",			TYPE_BINAIRE,	&p.isSauvegarde2},
	{"COUL_REMPLISSAGE",	TYPE_COULEUR,	&p.coul_remplissage}
};

void GetParams(int argc, _TCHAR* argv[])
{
	FILE	*pFile;
	char	pTmp[256];
	char	pTmp2[256];
	char	pConfFile[256];
	char	*pPtr;
	u32		qt_params = sizeof(g_pCmd) / sizeof(g_pCmd[0]);
	u32		i, r, g, b;
	float	tmp;
	TColor	*pCoul;

	// valeurs par dfaut
	p.d_bits = 0;
	p.file[0] = 0;
	strcpy(pConfFile, "config.txt");
	p.detect_width = 20;
	p.detect_diff = 25;
	p.vision_limit1 = 64;
	p.vision_limit2 = 3;
	p.vision_width1 = 16;
	p.vision_width2 = 16;
	p.vision_width3 = 128;
	p.vision_angle = 3.0;
	p.blur_radius = 3;
	p.blur_limit = 50;
	strcpy(p.repertoire1, "c:/t/");
	strcpy(p.repertoire2, "c:/trot/");
	p.cote_fond = BIT_DROITE;
	p.cote_recadrage = BIT_DROITE|BIT_BAS;
	p.isRecadrage1 = TRUE;
	p.isFlou = TRUE;
	p.isSauvegarde1 = TRUE;
	p.isRotation = TRUE;
	p.isRecadrage2 = TRUE;
	p.isSauvegarde2 = TRUE;
	p.isShowFilters = FALSE;
	p.isSaveBlacknWhite = FALSE;
	p.offset_angle = 0.0;
	p.isAutoColor = TRUE;

	// lit les paramtres de la ligne de commande
	for (i=1; i<(u32)argc; i++)
	{
		if (argv[i][0] == '/')
		{
			if (argv[i][1] == 'c')
			{
				GetRep(&argv[i][2], pConfFile);
			}
			else if (argv[i][1] == 'f')
			{
				p.isShowFilters = TRUE;
			}
			else if (argv[i][1] == 'b')
			{
				p.isSaveBlacknWhite = TRUE;
			}
			else if (argv[i][1] == 'd')
			{
				sscanf((char *)&argv[i][2], "%d", &p.d_bits);
			}
			else
			{
				Error("parametre inconnu: \"%s\"\n", argv[i]);
				p.file[0] = 0;
				return;
			}
		}
		else
		{
			GetRep(argv[i], p.file);
		}
	}

	// lit le fichier de config
	pFile = fopen(pConfFile, "r");

	if (pFile == NULL)
	{
		return;
	}

	while (fgets(pTmp, 256, pFile) != NULL)
	{
		while(pTmp[strlen(pTmp)-1] == 0xA || pTmp[strlen(pTmp)-1] ==0xD)
		{
			pTmp[strlen(pTmp)-1] = 0;
		}

		pPtr = strchr(pTmp, '=');

		if (pPtr != NULL)
		{
			memcpy(pTmp2, pTmp, pPtr-pTmp);
			pTmp2[pPtr-pTmp] = 0;
			strupr(pTmp2);

			for (i=0; i<qt_params; i++)
			{
				if (strcmp(pTmp2, g_pCmd[i].pCmd) == 0)
				{
					switch (g_pCmd[i].type)
					{
						case TYPE_INT:
							sscanf(pPtr+1, "%d", g_pCmd[i].pVar);
							break;

						case TYPE_FLOAT:
							tmp = 0.0;
							sscanf(pPtr+1, "%f", &tmp);
							*((double *)g_pCmd[i].pVar) = tmp;
							break;

						case TYPE_STRING:
							GetRep(pPtr+1, (char *)g_pCmd[i].pVar);
							break;

						case TYPE_COTE:
							*((u8 *)g_pCmd[i].pVar) = 0;
							strcpy(pTmp, pPtr+1);
							strupr(pTmp);
							pPtr = pTmp;

							while (*pPtr != 0)
							{
								switch (*pPtr)
								{
									case 'H':
										*((u8 *)g_pCmd[i].pVar) |= BIT_HAUT;
										break;

									case 'B':
										*((u8 *)g_pCmd[i].pVar) |= BIT_BAS;
										break;

									case 'G':
										*((u8 *)g_pCmd[i].pVar) |= BIT_GAUCHE;
										break;

									case 'D':
										*((u8 *)g_pCmd[i].pVar) |= BIT_DROITE;
										break;
								}
								pPtr++;
							}
							break;

						case TYPE_BINAIRE:
							if (pPtr[1] == 'o' || pPtr[1] == 'O')
							{
								*((u8 *)g_pCmd[i].pVar) = TRUE;
							}
							else
							{
								*((u8 *)g_pCmd[i].pVar) = FALSE;
							}
							break;

						case TYPE_COULEUR:
							p.isAutoColor = FALSE;
							sscanf(pPtr+1, "%d,%d,%d", &r, &g, &b);
							pCoul = (TColor *)g_pCmd[i].pVar;
							pCoul->r = (u8)r;
							pCoul->g = (u8)g;
							pCoul->b = (u8)b;
							break;
					}
					break;
				}
			}

			if (i == qt_params)
			{
				printf("parametre inconnu \"%s\"\n", pTmp2);
			}
		}
	}
	fclose(pFile);
}

/******************************************************************************
 *
 * Lit un pixel d'une bitmap
 *
 ******************************************************************************/
void GetPx(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	switch (p_pBm->format)
	{
		case BITMAP_GREY:
			pAddr = BITMAP_ADDR8(p_pBm, p_x, p_y);
			p_pCol->r = *pAddr;
			p_pCol->g = *pAddr;
			p_pCol->b = *pAddr;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
			p_pCol->r = *pAddr++;
			p_pCol->g = *pAddr++;
			p_pCol->b = *pAddr;
			break;
	}
}

/******************************************************************************
 *
 * Ecrit un pixel dans une bitmap
 *
 ******************************************************************************/
void SetPx(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	switch (p_pBm->format)
	{
		case BITMAP_GREY:
			*BITMAP_ADDR8(p_pBm, p_x, p_y) = p_pCol->r;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
			*pAddr++ = p_pCol->r;
			*pAddr++ = p_pCol->g;
			*pAddr   = p_pCol->b;
			break;
	}
}

/******************************************************************************
 *
 * Lit un pixel d'une bitmap version optimise pour les bitmaps en 24 bits
 *
 ******************************************************************************/
void GetPx24(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
	p_pCol->r = *pAddr++;
	p_pCol->g = *pAddr++;
	p_pCol->b = *pAddr;
}

/******************************************************************************
 *
 * Ecrit un pixel dans une bitmap version optimise pour les bitmaps en 24 bits
 *
 ******************************************************************************/
void SetPx24(TBitmap *p_pBm, int p_x, int p_y, TColor *p_pCol)
{
	u8	*pAddr;

	pAddr = BITMAP_ADDR24(p_pBm, p_x, p_y);
	*pAddr++ = p_pCol->r;
	*pAddr++ = p_pCol->g;
	*pAddr   = p_pCol->b;
}

/******************************************************************************
 *
 * Alloue une bitmap et sa mmoire
 *
 ******************************************************************************/
u32 AllocBm(TBitmap **p_pBitmap, u16 p_width, u16 p_height, u8 p_format)
{
	*p_pBitmap = (TBitmap *)MEM_Calloc(sizeof(TBitmap));

	if (*p_pBitmap == NULL)
	{
		return (FALSE);
	}

	if (BITMAP_Alloc(*p_pBitmap, p_width, p_height, p_format) == FALSE)
	{
		MEM_Free((void **)p_pBitmap);
		return (FALSE);
	}
	return (TRUE);
}

/******************************************************************************
 *
 * Remplace une bitmap par une autre en librant l'ancienne
 *
 ******************************************************************************/
void ReplaceBm(TBitmap **p_pOldBM, TBitmap *p_pNewBM)
{
	BITMAP_Free(*p_pOldBM);
	MEM_Free((void **)p_pOldBM);
	*p_pOldBM = p_pNewBM;
}

/******************************************************************************
 *
 * Redimensionne une Bitmap (ne marche pas avec les images en 256 couleurs!)
 *
 ******************************************************************************/
u32 Crop(TBitmap **p_pBm, s32 p_x1, s32 p_y1, s32 p_x2, s32 p_y2)
{
	TBitmap *pTmpBM;
	u16	sizex = (u16)(p_x2 - p_x1 + 1);
	u16	sizey = (u16)(p_y2 - p_y1 + 1);
	u32	y;

	if (AllocBm(&pTmpBM, sizex, sizey, (*p_pBm)->format) == FALSE)
	{
		return (FALSE);
	}

	if ((*p_pBm)->format == BITMAP_GREY)
	{
		for (y=0; y<sizey; y++)
		{
			memcpy(BITMAP_ADDR8(pTmpBM, 0, y), BITMAP_ADDR8(*p_pBm, p_x1, p_y1 + y), sizex);
		}
	}
	else
	{
		for (y=0; y<sizey; y++)
		{
			memcpy(BITMAP_ADDR24(pTmpBM, 0, y), BITMAP_ADDR24(*p_pBm, p_x1, p_y1 + y), sizex*3);
		}
	}

	ReplaceBm(p_pBm, pTmpBM);
	return (TRUE);
}

/******************************************************************************
 *
 * Trouve la dernire colonne de la page
 *
 ******************************************************************************/
s32 RecadreDroite(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (x=p_pBm->width-1; x>=0; x--)
	{
		for (y=0; y<p_pBm->height-p.detect_width; y++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx24(p_pBm, x, y+i, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (y != p_pBm->height-p.detect_width)
		{
			break;
		}
	}
	return (x);
}

/******************************************************************************
 *
 * Trouve la premire colonne de la page
 *
 ******************************************************************************/
s32 RecadreGauche(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (x=0; x<p_pBm->width; x++)
	{
		for (y=0; y<p_pBm->height-p.detect_width; y++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx24(p_pBm, x, y+i, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (y != p_pBm->height-p.detect_width)
		{
			break;
		}
	}
	return (x);
}

/******************************************************************************
 *
 * Trouve la dernire ligne de la page
 *
 ******************************************************************************/
s32 RecadreBas(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (y=p_pBm->height-1; y>=0; y--)
	{
		for (x=0; x<p_pBm->width-p.detect_width; x++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx24(p_pBm, x+i, y, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (x != p_pBm->width-p.detect_width)
		{
			break;
		}
	}
	return (y);
}

/******************************************************************************
 *
 * Trouve la premire ligne de la page
 *
 ******************************************************************************/
s32 RecadreHaut(TBitmap *p_pBm, TColor *p_pBcol)
{
	s32	x, y, i, dif;
	TColor	col;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width-p.detect_width; x++)
		{
			for (i=0; i<p.detect_width; i++)
			{
				GetPx24(p_pBm, x+i, y, &col);
				dif = abs((s32)col.r - (s32)p_pBcol->r) +
				      abs((s32)col.g - (s32)p_pBcol->g) +
				      abs((s32)col.b - (s32)p_pBcol->b);

				if (dif < p.detect_diff*3)
				{
					break;
				}
			}

			if (i == p.detect_width)
			{
				break;
			}
		}

		if (x != p_pBm->width-p.detect_width)
		{
			break;
		}
	}
	return (y);
}

/******************************************************************************
 *
 * Recadre dans tous les sens
 *
 ******************************************************************************/
u32 Recadre(TBitmap **p_pBm, TColor *p_pBcol)
{
	s32	x1, y1, x2, y2;

	x1 = 0;
	y1 = 0;
	x2 = (*p_pBm)->width - 1;
	y2 = (*p_pBm)->height - 1;

	if (p.cote_recadrage & BIT_HAUT)
	{
		y1 = RecadreHaut(*p_pBm, p_pBcol);
	}

	if (p.cote_recadrage & BIT_BAS)
	{
		y2 = RecadreBas(*p_pBm, p_pBcol);
	}

	if (p.cote_recadrage & BIT_GAUCHE)
	{
		x1 = RecadreGauche(*p_pBm, p_pBcol);
	}

	if (p.cote_recadrage & BIT_DROITE)
	{
		x2 = RecadreDroite(*p_pBm, p_pBcol);
	}

	return (Crop(p_pBm, x1, y1, x2, y2));
}

/******************************************************************************
 *
 * Trouve la couleur de fond de l'image
 *
 ******************************************************************************/
void GetBGCol(TBitmap *p_pBm, TColor *p_pBcol)
{
	TColor col;
	s32	i, r, g, b, s, c;

	r = 0;
	g = 0;
	b = 0;

	if (p.cote_fond & (BIT_HAUT|BIT_BAS))
	{
		s = p_pBm->width;

		for (i=0; i<s; i++)
		{
			c = (p.cote_fond & BIT_HAUT ? 0 : p_pBm->height-1);
			GetPx24(p_pBm, i, c, &col);
			r += col.r;
			g += col.g;
			b += col.b;
		}
	}
	else
	{
		s = p_pBm->height;

		for (i=0; i<s; i++)
		{
			c = (p.cote_fond & BIT_GAUCHE ? 0 : p_pBm->width-1);
			GetPx24(p_pBm, c, i, &col);
			r += col.r;
			g += col.g;
			b += col.b;
		}
	}

	p_pBcol->r = (u8)(r / s);
	p_pBcol->g = (u8)(g / s);
	p_pBcol->b = (u8)(b / s);
}

/******************************************************************************
 *
 * Initialise le noyau du flou gaussien
 *
 ******************************************************************************/
s32 *Gaussian_Init(void)
{
	s32	*pBlur = (s32 *)MEM_Alloc(sizeof(int)*(2*p.blur_radius+1)*(2*p.blur_radius+1));
	s32	*pBlur2 = pBlur;
	double sigma = (double)p.blur_radius / 3.0;
	double e;
	int	x, y;

	if (pBlur != NULL)
	{
		for (y=-p.blur_radius; y<p.blur_radius+1; y++)
		{
			for (x=-p.blur_radius; x<p.blur_radius+1; x++)
			{
				e = (double)(x*x + y*y);
				e = exp(-e / (2.0*sigma*sigma));
				*pBlur2++ = (int)(e * 1024.0);
			}
		}
	}

	return (pBlur);
}

/******************************************************************************
 *
 * Applique le flou gaussien
 *
 ******************************************************************************/
typedef struct
{
	s32		start_y, end_y;
} TGThread;

typedef struct
{
	s32		*pBlur;
	TBitmap	*pTmpBM;
	TBitmap	**pBM;
} TCommon_gaus;

TCommon_gaus g_cg;

long Gaussian_Thread(long param)
{
	TGThread	*par = (TGThread *)param;
	s32	x, y, x2, y2;
	s32	sum_r, sum_g, sum_b, sum_coef;
	s32	dist_r, dist_g, dist_b;
	s32	*pBlur2;
	TColor	col;

	// ces variables sont l pour acclrer le code
	s32	mid_r, mid_g, mid_b;
	s32	tmp_r, tmp_g, tmp_b;
	s32	limit = p.blur_limit*p.blur_limit;
	u8	*pPtr;
	s32	xmin, xmax;

	for (y=par->start_y; y<par->end_y; y++)
	{
		for (x=0; x<g_cg.pTmpBM->width; x++)
		{
			GetPx24(*(g_cg.pBM), x, y, &col);
			// tranformation de la couleur en s32 pour gagner du temps
			mid_r = col.r;
			mid_g = col.g;
			mid_b = col.b;
			sum_r = 0;
			sum_g = 0;
			sum_b = 0;
			sum_coef = 0;

			// calcule les limites en x ici pour viter un if  l'intrieur de la boucle
			xmin = x - p.blur_radius;

			if (xmin < 0)
			{
				xmin = 0;
			}
			xmax = x + p.blur_radius+1;

			if (xmax > g_cg.pTmpBM->width)
			{
				xmax = g_cg.pTmpBM->width;
			}

			for (y2=-p.blur_radius; y2<p.blur_radius+1; y2++)
			{
				if (y+y2 >= 0 && y+y2 < g_cg.pTmpBM->height)
				{
					// la bonne adresse dans la matrice du flou
					pBlur2 = &g_cg.pBlur[(2*p.blur_radius+1)*(y2+p.blur_radius)+xmin-x+p.blur_radius];

					// la bonne adresse du pixel
					pPtr = BITMAP_ADDR24(*(g_cg.pBM), xmin, y+y2);

					for (x2=xmin; x2<xmax; x2++)
					{
						// lit le pixel
						tmp_r = pPtr[0];
						tmp_g = pPtr[1];
						tmp_b = pPtr[2];
						dist_r = mid_r - tmp_r;
						dist_g = mid_g - tmp_g;
						dist_b = mid_b - tmp_b;
						// distance au carr (en dehors du if pour gagner du temps)
						dist_r *= dist_r;
						dist_g *= dist_g;
						dist_b *= dist_b;

						if (dist_r + dist_g + dist_b < limit)
						{
							sum_r += *pBlur2 * tmp_r;
							sum_g += *pBlur2 * tmp_g;
							sum_b += *pBlur2 * tmp_b;
							sum_coef += *pBlur2;
						}
						pBlur2++;
						pPtr += 3;
					}
				}
			}
			col.r = (u8)(sum_r / sum_coef);
			col.g = (u8)(sum_g / sum_coef);
			col.b = (u8)(sum_b / sum_coef);
			SetPx24(g_cg.pTmpBM, x, y, &col);
		}
	}
	return(0);
}

u32 Gaussian_Process(TBitmap **p_pBm)
{
	TGThread	prm[NUM_THREAD];
	HANDLE	h[NUM_THREAD-1];
	DWORD	id;
	u32		i;

	// init du flou
	g_cg.pBlur = Gaussian_Init();

	if (g_cg.pBlur == NULL)
	{
		return(FALSE);
	}

	if (AllocBm(&g_cg.pTmpBM, (*p_pBm)->width, (*p_pBm)->height, (*p_pBm)->format) == FALSE)
	{
		MEM_Free((void **)&g_cg.pBlur);
		return(FALSE);
	}

	g_cg.pBM = p_pBm;

	for (i=0; i<NUM_THREAD; i++)
	{
		prm[i].start_y =   (i * (*p_pBm)->height) / NUM_THREAD;
		prm[i].end_y = ((i+1) * (*p_pBm)->height) / NUM_THREAD;
	}

	for (i=0; i<NUM_THREAD-1; i++)
	{
		h[i] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)Gaussian_Thread, (LPVOID)&prm[i], 0, &id);
	}
	Gaussian_Thread((long)&prm[NUM_THREAD-1]);

	WaitForMultipleObjects(NUM_THREAD-1, h, TRUE, INFINITE);

	for (i=0; i<NUM_THREAD-1; i++)
	{
		CloseHandle(h[i]);
	}

	MEM_Free((void **)&g_cg.pBlur);
	ReplaceBm(p_pBm, g_cg.pTmpBM);
	return (TRUE);
}

/******************************************************************************
 *
 * Filtre 1: ne garde que les transitions verticales franches
 *
 ******************************************************************************/
void Filtre1(TBitmap *p_pBm)
{
	s32	x, y, c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			if (y == p_pBm->height-1)
			{
				*pPtr++ = 0;
			}
			else
			{
				c = abs((s32)pPtr[0] - (s32)pPtr[p_pBm->width]);
				*pPtr++ = (c < p.vision_limit1 ? 0 : 255);
			}
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE && p.d_bits == 0)
	{
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre1.tga");
		TGA_Save(pTmpFilename, p_pBm);
	}

	// sauvegarde pour retouche_gui
	dsave(4, p_pBm, FALSE);
}

/******************************************************************************
 *
 * Filtre 2: ne garde que les transition frquentes
 *
 ******************************************************************************/
void Filtre2(TBitmap *p_pBm)
{
	s32	x, y, x2, x3, c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			c = 0;

			if (x < p_pBm->width-p.vision_width1)
			{
				for (x2=0; x2<p.vision_width1-p.vision_limit2; x2++)
				{
					for (x3=0; x3<p.vision_limit2; x3++)
					{
						if (pPtr[x2+x3] == 0)
						{
							break;
						}
					}

					if (x3 == p.vision_limit2)
					{
						c++;
					}
				}
			}
			*pPtr++ = (c != 0 ? 255 : 0);
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE && p.d_bits == 0)
	{
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre2.tga");
		TGA_Save(pTmpFilename, p_pBm);
	}

	// sauvegarde pour retouche_gui
	dsave(5, p_pBm, FALSE);
}

/******************************************************************************
 *
 * Filtre 3: tend les lignes
 *
 ******************************************************************************/
void Filtre3(TBitmap *p_pBm)
{
	s32	x, y, x2, x3;
	u8	*pPtr;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			pPtr = BITMAP_ADDR8(p_pBm, x, y);

			// le pixel point par pPtr est noir ?
			if (pPtr[0] == 0 && x != 0)
			{
				// le pixel  gauche est blanc ? (cd qu'on est au bout d'une ligne)
				if (pPtr[-1] == 255)
				{
					// on continue vers la droite jusqu' la longueur max...
					for (x2=0; x2<p.vision_width2; x2++)
					{
						// on est arriv au bout de l'image ?
						if (x+x2 >= p_pBm->width)
						{
							x = p_pBm->width;
							break;
						}

						// on a trouv le dbut d'une autre ligne sur la mme ligne, sur la ligne au-dessus ou sur la ligne en-dessous ?
						if (pPtr[x2] == 255 || (y != 0 && pPtr[x2-p_pBm->width] == 255) || (y != p_pBm->height-1 && pPtr[x2+p_pBm->width] == 255))
						{
							// si oui, on crit des "1"
							for (x3=0; x3<x2; x3++)
							{
								*pPtr++ = 1;
								x++;
							}
							pPtr++;
							x++;
							break;
						}
					}
				}
			}
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE || p.d_bits != 0)
	{
		TBitmap	bm;
		TColor	c;

		BITMAP_Alloc(&bm, p_pBm->width, p_pBm->height, BITMAP_GREY);

		for (y=0; y<bm.height; y++)
		{
			for (x=0; x<bm.width; x++)
			{
				GetPx(p_pBm, x, y, &c);

				if (c.r == 1)
				{
					// les pixels qui taient  1 deviennent gris
					c.r = c.g = c.b = 128;
				}
				SetPx(&bm, x, y, &c);
			}
		}

		if (p.d_bits == 0)
		{
			strcpy(pTmpFilename, p.repertoire2);
			strcat(pTmpFilename, "retouche_filtre3.tga");
			TGA_Save(pTmpFilename, &bm);
		}
		else
		{
			// sauvegarde pour retouche_gui
			dsave(6, &bm, FALSE);
		}
		BITMAP_Free(&bm);
	}
}

/******************************************************************************
 *
 * Filtre 4: lignes d'1 pixel d'paisseur
 *
 ******************************************************************************/
void Filtre4(TBitmap *p_pBm)
{
	s32	x, y;
	u8	c;
	u8	*pPtr;

	pPtr = p_pBm->pMem;

	for (y=0; y<p_pBm->height; y++)
	{
		for (x=0; x<p_pBm->width; x++)
		{
			c = 0;

			if (y < p_pBm->height-1)
			{
				if (pPtr[0] != 0 && pPtr[p_pBm->width] == 0)
				{
					c = 255;
				}
			}
			*pPtr++ = c;
		}
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE && p.d_bits == 0)
	{
		strcpy(pTmpFilename, p.repertoire2);
		strcat(pTmpFilename, "retouche_filtre4.tga");
		TGA_Save(pTmpFilename, p_pBm);
	}

	// sauvegarde pour retouche_gui
	dsave(7, p_pBm, FALSE);
}

/******************************************************************************
 *
 * Filtre 5: moyenne des angles des lignes les plus longues
 *
 ******************************************************************************/
double Filtre5(TBitmap *p_pBm)
{
	s32	x, y, x2, y2;
	double angle = 0.0;
	double angle2, dist;
	double fac = 0.0;

	// On parcourt l'image de gauche  droite, puis de haut en bas pour trouver tous les dbuts de lignes
	for (x=0; x<p_pBm->width; x++)
	{
		for (y=0; y<p_pBm->height; y++)
		{
			if (*BITMAP_ADDR8(p_pBm, x, y) == 255)
			{
				// on a trouv le dbut d'une ligne
				x2 = x;
				y2 = y;

				// on la suit pour voir o elle s'arrte
				while (1)
				{
					// on colorie cette ligne temporairement avec la couleur 1
					*BITMAP_ADDR8(p_pBm, x2, y2) = 1;

					if (x2 >= p_pBm->width-1)
					{
						break;
					}

					if (*BITMAP_ADDR8(p_pBm, x2+1, y2) > 128)
					{
						// essaye d'aller  droite
						x2++;
					}
					else if (y2 > 0 && *BITMAP_ADDR8(p_pBm, x2+1, y2-1) > 128)
					{
						// essaye d'aller en haut  droite
						x2++;
						y2--;
					}
					else if (y2 < p_pBm->height-1 && *BITMAP_ADDR8(p_pBm, x2+1, y2+1) > 128)
					{
						// essaye d'aller en bas  droite
						x2++;
						y2++;
					}
					else
					{
						break;
					}
				}

				// est-ce que la ligne est assez longue
				if (x2 - x >= p.vision_width3)
				{
					angle2 = atan2((double)y - (double)y2, (double)x2 - (double)x);
					angle2 = angle2 * 180.0 / pi;

					// est-ce que l'angle de la ligne n'est pas trop grand ?
					if (fabs(angle2) < p.vision_angle)
					{
						dist = (double)(x2-x);
						angle += dist * angle2;
						fac += dist;

						// si l'on veut sauver le rsultat, on recolorie cette ligne
						if (p.isShowFilters == TRUE || p.d_bits != 0)
						{
							x2 = x;
							y2 = y;

							while (1)
							{
								*BITMAP_ADDR8(p_pBm, x2, y2) = 2;

								if (x2 >= p_pBm->width-1)
								{
									break;
								}

								if (*BITMAP_ADDR8(p_pBm, x2+1, y2) == 1)
								{
									x2++;
								}
								else if (y2 > 0 && *BITMAP_ADDR8(p_pBm, x2+1, y2-1) == 1)
								{
									x2++;
									y2--;
								}
								else if (y2 < p_pBm->height-1 && *BITMAP_ADDR8(p_pBm, x2+1, y2+1) == 1)
								{
									x2++;
									y2++;
								}
								else
								{
									break;
								}
							}
						}
					}
				}
			}
		}
	}

	if (fac != 0.0)
	{
		angle = angle / fac;
	}

	// sauvegarde du filtre pour le rglage
	if (p.isShowFilters == TRUE || p.d_bits != 0)
	{
		TBitmap	bm;
		TColor	c;

		BITMAP_Alloc(&bm, p_pBm->width, p_pBm->height, BITMAP_GREY);

		for (y=0; y<bm.height; y++)
		{
			for (x=0; x<bm.width; x++)
			{
				GetPx(p_pBm, x, y, &c);

				if (c.r == 1)
				{
					// les pixels  1 se transforment en gris
					c.r = c.g = c.b = 128;
				}
				else if (c.r == 2)
				{
					// les pixels  2 se transforment en blanc
					c.r = c.g = c.b = 255;
				}
				SetPx(&bm, x, y, &c);
			}
		}

		if (p.d_bits == 0)
		{
			strcpy(pTmpFilename, p.repertoire2);
			strcat(pTmpFilename, "retouche_filtre5.tga");
			TGA_Save(pTmpFilename, &bm);
		}

		// sauvegarde pour retouche_gui
		dsave(8, &bm, FALSE);
		BITMAP_Free(&bm);
	}

	return (angle);
}

/******************************************************************************
 *
 * calcule un coefficient pour l'interpolation bicubique en fonction de la distance du point
 *
 ******************************************************************************/
double Rotate_CalcCoef(double p_dist)
{
	double	dist2 = p_dist * p_dist;

	if (p_dist < 1.0)
	{
		return (1.5 * dist2*p_dist - 2.5 * dist2 + 1.0);
	}
	else if (p_dist < 2.0)
	{
		return (-0.5 * dist2*p_dist + 2.5 * dist2 - 4.0 * p_dist + 2.0);
	}
	else
	{
		return (0.0);
	}
}

/******************************************************************************
 *
 * limite une valeur entre 0 et 255 avec un arrondi
 *
 ******************************************************************************/
u8 limit(double p_val)
{
	double f = floor(p_val);
	double c = ceil(p_val);
	s32	i;

	if (c-p_val < p_val-f)
	{
		i = (s32)c;
	}
	else
	{
		i = (s32)f;
	}

	if (i < 0)
	{
		return (0);
	}
	else if (i > 255)
	{
		return (255);
	}
	else
	{
		return ((u8)i);
	}
}

/******************************************************************************
 *
 * trouve l'image d'un point aprs la rotation, en utilisant une interpolation bicubique
 *
 ******************************************************************************/
void RotatePoint(TBitmap *p_pSrc, double p_x, double p_y, TColor *p_pBcol, TColor *p_pCol)
{
	s32	intx = (int)floor(p_x);
	s32	inty = (int)floor(p_y);
	s32	srcx, srcy;
	double	dist, coef,coefy;
	double	fr, fg, fb, fac;
	u8	*pPtr;
	double pCoefx[4];

	// prcalcule les coefficients en x
	for (srcx=intx-1; srcx<intx+3; srcx++)
	{
		dist = (double)srcx + 0.5 - p_x;
		pCoefx[srcx - intx+1] = Rotate_CalcCoef(fabs(dist));
	}

	fac = 0.0;
	fr = 0.0;
	fg = 0.0;
	fb = 0.0;

	for (srcy=inty-1; srcy<inty+3; srcy++)
	{
		dist = fabs((double)srcy + 0.5 - p_y);
		coefy = Rotate_CalcCoef(dist);
		pPtr = BITMAP_ADDR24(p_pSrc, intx-1, srcy);

		for (srcx=intx-1; srcx<intx+3; srcx++)
		{
			coef = coefy * pCoefx[srcx - intx+1];
			fac += coef;

			if (srcx >= 0 && srcx < p_pSrc->width &&
				srcy >= 0 && srcy < p_pSrc->height)
			{
				fr += coef * (double)pPtr[0];
				fg += coef * (double)pPtr[1];
				fb += coef * (double)pPtr[2];
			}
			else
			{
				fr += coef * (double)p_pBcol->b;
				fg += coef * (double)p_pBcol->g;
				fb += coef * (double)p_pBcol->r;
			}
			pPtr += 3;
		}
	}
	p_pCol->r = limit(fr / fac);
	p_pCol->g = limit(fg / fac);
	p_pCol->b = limit(fb / fac);
}



typedef struct
{
	s32		start_y, end_y;
	double	savex, savey;
} TThread;

typedef struct
{
	double	incx, incy;
	TBitmap	*pSrc;
	TBitmap	*pDst;
	TColor	*pBcol;
} TCommon_rot;

TCommon_rot g_cr;

/******************************************************************************
 *
 * Rotation de l'image
 *
 ******************************************************************************/

long Rotate_Thread(long param)
{
	TThread	*par = (TThread *)param;
	double	src_x, src_y;
	s32		dst_x, dst_y;
	TColor	col;

	par->savex -= par->start_y * g_cr.incy;
	par->savey += par->start_y * g_cr.incx;

	for (dst_y=par->start_y; dst_y<par->end_y; dst_y++)
	{
		src_x = par->savex;
		src_y = par->savey;
		par->savex -= g_cr.incy;
		par->savey += g_cr.incx;

		for (dst_x=0; dst_x<g_cr.pDst->width; dst_x++)
		{
			RotatePoint(g_cr.pSrc, src_x, src_y, g_cr.pBcol, &col);
			SetPx24(g_cr.pDst, dst_x, dst_y, &col);
			src_x += g_cr.incx;
			src_y += g_cr.incy;
		}
	}
	return (0);
}

void Rotate(TBitmap *p_pSrc, TBitmap *p_pDst, double p_angle, TColor *p_pBcol)
{
	TThread	prm[NUM_THREAD];
	HANDLE	h[NUM_THREAD-1];
	double	center_x, center_y;
	double	angle, angle2;
	double	dist;
	double	savex, savey;
	u32		i;
	DWORD	id;

	g_cr.pSrc = p_pSrc;
	g_cr.pDst = p_pDst;
	g_cr.pBcol = p_pBcol;

	angle = p_angle * pi / 180.0;
	g_cr.incx = cos(angle);
	g_cr.incy = -sin(angle);

	center_x = (float)p_pSrc->width / 2.0 - 0.5;
	center_y = (float)p_pSrc->height / 2.0 - 0.5;
	dist = sqrt(center_x*center_x + center_y*center_y);
	angle2 = atan2(-center_y, center_x);
	savex = center_x+0.5 - dist * cos(angle2+angle);
	savey = center_y+0.5 + dist * sin(angle2+angle);

	for (i=0; i<NUM_THREAD; i++)
	{
		prm[i].start_y =   (i * p_pDst->height) / NUM_THREAD;
		prm[i].end_y = ((i+1) * p_pDst->height) / NUM_THREAD;
		prm[i].savex = savex;
		prm[i].savey = savey;
	}

	for (i=0; i<NUM_THREAD-1; i++)
	{
		h[i] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)Rotate_Thread, (LPVOID)&prm[i], 0, &id);
	}
	Rotate_Thread((long)&prm[NUM_THREAD-1]);

	WaitForMultipleObjects(NUM_THREAD-1, h, TRUE, INFINITE);

	for (i=0; i<NUM_THREAD-1; i++)
	{
		CloseHandle(h[i]);
	}
}

/******************************************************************************
 *
 * Dtection de l'angle et rotation
 *
 ******************************************************************************/
u32 Autorotate(TBitmap **p_pBm, TColor *p_pBack_col)
{
	double	angle;
	TBitmap work_bitmap;
	TBitmap	*pTmpBM;

	// cre le buffer de travail pour la dtection de l'angle
	if (BITMAP_Alloc(&work_bitmap, (*p_pBm)->width, (*p_pBm)->height, BITMAP_GREY) == FALSE)
	{
		return (FALSE);
	}
	BITMAP_Blt(*p_pBm, &work_bitmap, 0, 0);

	// ne garde que les transitions verticales franches
	Filtre1(&work_bitmap);

	// ne garde que les transition frquentes
	Filtre2(&work_bitmap);

	// tend les lignes
	Filtre3(&work_bitmap);

	// lignes d'1 pixel
	Filtre4(&work_bitmap);

	// moyenne des angles des lignes les plus longues
	angle = Filtre5(&work_bitmap);
	printf("%f\n", angle);
	BITMAP_Free(&work_bitmap);

	if (p.isRotation == TRUE && chk_bits(9))
	{
		// ajoute l'offset
		angle += p.offset_angle;

		// rotation
		if (AllocBm(&pTmpBM, (*p_pBm)->width, (*p_pBm)->height, (*p_pBm)->format) == FALSE)
		{
			return (FALSE);
		}

		BITMAP_Clr(pTmpBM, p_pBack_col->r, p_pBack_col->g, p_pBack_col->b);
		Rotate(*p_pBm, pTmpBM, angle, (p.isAutoColor == FALSE ? &p.coul_remplissage : p_pBack_col));
		ReplaceBm(p_pBm, pTmpBM);
	}

	// sauvegarde pour retouche_gui
	dsave(9, *p_pBm, FALSE);
	return (TRUE);
}

/******************************************************************************
 *
 * Convertit une bitmap dans un autre format
 *
 ******************************************************************************/
u32 ConvertBm(TBitmap **p_pBm, u8 p_format)
{
	TBitmap	*pTmpBM;

	if (AllocBm(&pTmpBM, (*p_pBm)->width, (*p_pBm)->height, p_format) == FALSE)
	{
		return (FALSE);
	}
	BITMAP_Blt(*p_pBm, pTmpBM, 0, 0);
	ReplaceBm(p_pBm, pTmpBM);
	return (TRUE);
}

/******************************************************************************
 *
 * Sauve l'image en la convertissant si besoin
 *
 ******************************************************************************/
void SaveImage(TBitmap *p_pBm, char *p_pRep, u32 p_isGrey)
{
	char	pFilename[256];
	TBitmap	save_bm;
	char	*pTmp;

	strcpy(pFilename, p_pRep);

	if (p_pRep[strlen(p_pRep)-1] != '\\' && p_pRep[strlen(p_pRep)-1] != '/')
	{
		strcat(pFilename, "\\");
	}
	strcat(pFilename, p.file);
	pTmp = strrchr(pFilename, '.');

	if (p.isSaveBlacknWhite == FALSE)
	{
		strcpy(pTmp, ".tga");
	}
	else
	{
		strcpy(pTmp, ".tif");
	}

	if (p_isGrey == FALSE && p.isSaveBlacknWhite == FALSE)
	{
		TGA_Save(pFilename, p_pBm);
	}
	else
	{
		BITMAP_Alloc(&save_bm, p_pBm->width, p_pBm->height, BITMAP_GREY);
		BITMAP_Blt(p_pBm, &save_bm, 0, 0);

		if (p.isSaveBlacknWhite == FALSE)
		{
			TGA_Save(pFilename, &save_bm);
		}
		else
		{
			TIFF_Save(pFilename, &save_bm);
		}
		BITMAP_Free(&save_bm);
	}
}

/******************************************************************************
 *
 * sauvegarde de chaque tape pour retouche_gui
 *
 ******************************************************************************/
void dsave(u32 p_bit, TBitmap *p_pBm, u32 p_convert)
{
	u32	x, y;
	char pFilename[MAX_PATH];

	if (p.d_bits & (1 << p_bit))
	{
		if (p.repertoire2[strlen(p.repertoire2)-1] != '\\' && p.repertoire2[strlen(p.repertoire2)-1] != '/')
		{
			strcat(p.repertoire2, "\\");
		}
		sprintf(pFilename, "%ssave%d.tga", p.repertoire2, p_bit);

		if (p_convert == FALSE || p.isSaveBlacknWhite == FALSE)
		{
			TGA_Save(pFilename, p_pBm);
		}
		else
		{
			TBitmap	bm;
			TColor	c;

			BITMAP_Alloc(&bm, p_pBm->width, p_pBm->height, BITMAP_GREY);

			for (y=0; y<bm.height; y++)
			{
				for (x=0; x<bm.width; x++)
				{
					GetPx(p_pBm, x, y, &c);

					if ((c.r*c.r + c.g*c.g + c.b*c.b) > 127*127)
					{
						c.r = c.g = c.b = 255;
					}
					else
					{
						c.r = c.g = c.b = 0;
					}
					SetPx(&bm, x, y, &c);
				}
			}
			TGA_Save(pFilename, &bm);
			BITMAP_Free(&bm);
		}
	}
}

u32 chk_bits(u32 p_bit)
{
	if (p.d_bits == 0 || (p.d_bits & (1 << p_bit)))
	{
		return (TRUE);
	}
	return (FALSE);
}

/******************************************************************************
 *
 * Point d'entre du programme
 *
 ******************************************************************************/
int _tmain(int argc, _TCHAR* argv[])
{
	u32		isFileLoaded;
	u32		ret;
	TBitmap	*pBm = NULL;
	u32		isGrey = FALSE;
	TColor	back_col;

	// init
	GetParams(argc, argv);

	if (p.file[0] == 0)
	{
		printf("\n");
		printf("retouche v3.0 par Frederic Goset\n");
		printf("  usage: retouche [flags] <fichier image>\n");
		printf("\n");
		printf("les flags possibles sont:\n");
		printf("  /c<fichier de config> : le nom du fichier config a utiliser\n");
		printf("                          exemple: retouche /c\"config2.txt\" image1.tga\n");
		printf("  /f : pour sauver les resultats des filtres dans le repertoire 2\n");
		printf("  /b : pour sauver les images au format TIFF noir et blanc\n");
		printf("  /d<val> : sauvegarde des differentes etapes (pour retouche_gui)\n");
		return (0);
	}

	// alloue la bitmap de dpart
	pBm = (TBitmap *)MEM_Alloc(sizeof(TBitmap));

	// chargement
	isFileLoaded = TIFF_Load(p.file, pBm);

	if (isFileLoaded == FALSE)
	{
		isFileLoaded = TGA_Load(p.file, pBm);
	}

	if (isFileLoaded == TRUE)
	{
		// sauvegarde pour retouche_gui
		dsave(0, pBm, FALSE);

		SplitPath(p.file, p.srcdir);
		ret = TRUE;

		// conversion du format
		if (pBm->format == BITMAP_RGB)
		{
			// la routine de sauvegarde TGA n'accepte pas le format RGB: on le convertit en BGR
			ret = ConvertBm(&pBm, BITMAP_BGR);
		}
		else if (pBm->format == BITMAP_256)
		{
			// le programme n'est pas conu pour les images en 256 couleurs: on les convertit en tons de gris
			ret = ConvertBm(&pBm, BITMAP_GREY);
		}

		// a ce stade, l'image ne peut tre qu'en GREY ou en BGR.
		// si elle est en GREY, on la convertit en BGR pour optimiser les tapes suivantes
		if (pBm->format == BITMAP_GREY)
		{
			isGrey = TRUE;
			ret = ConvertBm(&pBm, BITMAP_BGR);
		}

		if (ret == TRUE)
		{
			// get background color
			GetBGCol(pBm, &back_col);

			// recadrage
			if (p.isRecadrage1 == TRUE && chk_bits(1))
			{
				ret = Recadre(&pBm, &back_col);
			}

			// sauvegarde pour retouche_gui
			dsave(1, pBm, FALSE);

			if (ret == TRUE)
			{
				// flou gaussien
				if (p.isFlou == TRUE && chk_bits(2))
				{
					ret = Gaussian_Process(&pBm);
				}

				// sauvegarde pour retouche_gui
				dsave(2, pBm, FALSE);

				if (ret == TRUE)
				{
					// sauvegarde 1
					if (p.isSauvegarde1 == TRUE && p.d_bits == 0)
					{
						SaveImage(pBm, p.repertoire1, isGrey);
					}

					// sauvegarde pour retouche_gui
					dsave(3, pBm, TRUE);

					// rotation
					if (p.isRotation == TRUE || p.isShowFilters == TRUE || p.d_bits != 0)
					{
						ret = Autorotate(&pBm, &back_col);
					}

					if (ret == TRUE)
					{
						// recadrage 2
						if (p.isRecadrage2 == TRUE && chk_bits(10))
						{
							ret = Recadre(&pBm, &back_col);
						}

						// sauvegarde pour retouche_gui
						dsave(10, pBm, FALSE);

						if (ret == TRUE)
						{
							if (p.isSauvegarde2 == TRUE && p.d_bits == 0)
							{
								// sauvegarde 2
								SaveImage(pBm, p.repertoire2, isGrey);
							}

							// sauvegarde pour retouche_gui
							dsave(11, pBm, TRUE);
							printf("%s done.\n", p.file);
						}
					}
				}
			}
		}
	}

	if (pBm != NULL)
	{
		BITMAP_Free(pBm);
		MEM_Free((void **)&pBm);
	}
	return 0;
}
